/****************************************************************************
** $Id: main.cpp  beta5   edited May 28 14:36 $
**
** Copyright (C) 2001-2002 Trolltech AS.  All rights reserved.
**
** This file is part of the Qt Script for Applications framework (QSA).
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** Licensees holding a valid QSA Beta Evaluation Version license may use
** this file in accordance with the QSA Beta Evaluation Version License
** Agreement provided with the Software.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/pricing.html or email sales@trolltech.com for
**   information about QSA Commercial License Agreements.
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
*****************************************************************************/

#include <qfile.h>
#include <qprocess.h>
#include <qdir.h>
#include <qstringlist.h>
#include <qregexp.h>
#include <qapplication.h>
#include <stdlib.h>
#ifdef Q_WS_WIN
#include <windows.h>
#endif

static QString *qtDir = 0;
static QString *assistant = 0;
static const char *app = 0;
static int processes = 0;
static int errors = 0;
static int warnings = 0;

static void message( const QString &msg )
{
    qWarning( msg );
}

class ProcessListener : public QObject {
    Q_OBJECT
public:
    ProcessListener( QProcess *pr ) : p( pr ) { processes++; }
    ~ProcessListener() {
	delete p;
	if ( --processes == 0 )
	    QApplication::exit();
    }

public slots:
    void processExited() {
	if( p->exitStatus()!=0 )
	    warnings++;
	deleteLater();
    }

    void readyReadStdout() {
	if( p->canReadLineStdout() ) {
	    qDebug( p->readLineStdout() );
	}
   }

    void readyReadStderr() {
	if( p->canReadLineStderr() ) {
	    qDebug(  p->readLineStderr() );
	}
    }

private:
    QProcess *p;
};


static bool execute( const QStringList &args )
{
    QProcess *p = new QProcess( args );
    ProcessListener *l = new ProcessListener( p );
    QObject::connect( p, SIGNAL( readyReadStdout() ), l, SLOT( readyReadStdout() ) );
    QObject::connect( p, SIGNAL( readyReadStderr() ), l, SLOT( readyReadStderr() ) );
    QObject::connect( p, SIGNAL( processExited() ), l, SLOT( processExited() ) );
    if( !p->start() ) {
	delete l;
	message( "Failed to start process: '" + args.join( " " ) + "'" );
	return FALSE;
    }
    return TRUE;
}

#if QT_VERSION < 0x030200
static void addContentFile( const QString &name )
{
    QStringList lst;
    lst.append( *assistant );
    // Assistant works differently in 3.2
    lst.append( "-addContentFile" );
    lst.append( QDir::convertSeparators( name ) );
    if( !execute( lst ) ) {
	warnings++;
    }
}
#endif

static void installDocs()
{
    if ( !assistant ) {
	assistant = new QString( "assistant" );
#ifdef Q_OS_MACX
	QStringList dirs = QStringList::split( ':', getenv( "PATH" ) );
	QStringList::const_iterator it;
	for ( it = dirs.begin(); it != dirs.end(); ++it ) {
	    QString a = *it + "/assistant.app/Contents/MacOS/assistant";
	    if ( QFile::exists( a ) ) {
		*assistant = a;
		break;
	    }
	}
#endif
    }
#if QT_VERSION < 0x030200
    // install documentation
    QString html = "doc/html/";
    addContentFile( html + "qsa.xml" );
    addContentFile( html + "qt-script-for-applications.xml" );
#else
    QStringList lst;
    lst.append( *assistant );
    lst.append( "-addProfile" );
    lst.append( QString( "doc" ) + QDir::separator() + QString( "docprofile.adp" ) );
    if( !execute( lst ) )
	warnings++;
#endif
}


static void mkDir( const QString &dir )
{
    QDir current;
    current.mkdir( dir );
}

static void copy( const QString &source, const QString &dest )
{
    QString s = QDir::convertSeparators( source );
    QString d = QDir::convertSeparators( dest );
#ifdef Q_OS_UNIX
    system( "cp " + QFile::encodeName( s ) + " " + QFile::encodeName( d ) );
    system( "chmod +w " + QFile::encodeName( d ) );
#else
    QT_WA(
    {
	if ( !CopyFileW( (TCHAR*) s.ucs2(), (TCHAR*) d.ucs2(), FALSE ) ) {
	    message( "Failed to copy file: " + s );
	    errors++;
	}
	if ( !SetFileAttributesW( (TCHAR*) d.ucs2(), FILE_ATTRIBUTE_NORMAL ) ) {
	    message( "Failed to set file attributes to normal" );
	    errors++;
	}
    }, {
	if ( !CopyFileA( QFile::encodeName( s ), QFile::encodeName( d ), FALSE ) ) {
	    message( "Failed to copy file: " + s );
	    errors++;
	}
	if ( !SetFileAttributesA( QFile::encodeName( d ), FILE_ATTRIBUTE_NORMAL ) ) {
	    message( "Failed to set file attributes to normal" );
	    errors++;
	}
    } );

#endif
}

static void symLink( const QString &source, const QString &dest )
{
#ifdef Q_OS_UNIX
    QString s = QDir::convertSeparators( source );
    QFileInfo info( s );
    s = info.absFilePath();
    QString d = QDir::convertSeparators( dest );
    system( "rm -f " + QFile::encodeName( d ) );
    system( "ln -s " + QFile::encodeName( s ) + " " + QFile::encodeName( d ) );
#else
    copy( source, dest );
#endif
}

static bool checkLicense()
{
    enum LicenseType { None, GPL, Evaluation, Professional, Enterprise, Internal };
    LicenseType qsaLicense = None;
    LicenseType qtLicense = None;

    if ( QFile::exists( *qtDir + "LICENSE.TROLL" ) ) {
	qtLicense = Internal;
    } else if( QFile::exists( *qtDir + "LICENSE.EVAL" ) ) {
	qtLicense = Evaluation;
    } else if ( QFile::exists( *qtDir + "LICENSE.GPL" ) ||
		!QFile::exists( *qtDir + "LICENSE" ) &&
		!QFile::exists( QDir::homeDirPath() + "/.qt-license" ) ) {
	qtLicense = GPL;
    } else if ( QFile::exists( *qtDir + "LICENSE" ) ) {
	QString home = QDir::homeDirPath();
	QString licenseFile = home + "/.qt-license";
	QFile f( licenseFile );
	if ( f.exists() ) {
	    if ( f.open( IO_ReadOnly ) ) {
		QString text = f.readAll();
		bool enterpriseQt = text.find( "\"qt-enterprise\"" ) >= 0;
		qtLicense = enterpriseQt ? Enterprise : Professional;
#if 0
		QRegExp re( "LicenseKey=([^\n]*)" );
		Q_ASSERT( re.isValid() );
		if ( re.search( text ) >= 0 ) {
		    QString key = re.cap( 1 );
		    if ( !key.isEmpty() ) {
			if ( QString( "3579BDFHbdfh" ).find( key[0] ) >= 0 )
			    usVersion = TRUE;
		    }
		}
#endif
	    }
	}
    } else {
	message( "\nCan't find license file of your Qt Edition\n" );
	return FALSE;
    }

    if ( QFile::exists( "LICENSE.GPL" ) ) {
	qsaLicense = GPL;
    } else if ( QFile::exists( "LICENSE.EVAL" ) ) {
	qsaLicense = Evaluation;
    }

    if ( qtLicense == Internal ||
	 qtLicense == qsaLicense ||
	 qtLicense == Enterprise && qsaLicense == Evaluation ) {
	return TRUE;
    } else if ( ( qtLicense == Enterprise || qtLicense == Professional ) &&
		qsaLicense == GPL ) {
	message( "\nThis QSA Free Edition is licensed under the GNU General "
		 "Public License.\nNote that any derivatives of this software "
		 "must also be licensed under\nthe GNU GPL.\n\n"
		 "See http://www.trolltech.com/gpl for more information or\n"
		 "contact sales@trolltech.com for an Evaluation Edition.\n" );
	return TRUE;
    } else {
	message( "\nThe required license information cannot be found"
		 "\nor the information is conflicting!\n" );
	QString str = QString( "Error code: %1%2\n\n" ).arg( qtLicense ).arg( qsaLicense );
	message( str );
	return FALSE;
    }
    return FALSE;
}

static bool writeQSConfig( bool buildIde )
{
    QFile file( "src/qsa/qsconfig.h" );
    if( !file.open( IO_WriteOnly ) ) {
	message( "Failed to open 'src/qsa/qsconfig.h' for writing." );
	return FALSE;
    }
    QTextStream txt( &file );
    txt << "// This file is autogenerated by QSA configure, do not modify it!\n"
	<< "#ifndef QS_CONFIG_H\n"
	<< "#define QS_CONFIG_H\n"
	<< "\n";
    if( !buildIde )
	txt << "#define QSA_NO_IDE\n";
    txt << "\n"
	<< "#endif\n";
    return TRUE;
}

int main( int argc, char **argv )
{
    app = argv[0];
    QApplication qapp( argc, argv, FALSE );

    QStringList configs;
    QString prefix;
    bool buildIde = TRUE;
    for ( int i = 1; i < qapp.argc(); i++ ) {
	QString arg = qapp.argv()[i];
	if ( arg.startsWith( "--" ) )
	    arg = arg.mid( 1 );
	if ( arg == "-help"  ) {
	    ; // handled by the configure script
	} else if ( arg == "-thread" ) {
	    configs << "thread";
	} else if ( arg == "-prefix" ) {
	    if ( i + 1 < qapp.argc() ) {
		prefix = qapp.argv()[++i];
	    } else {
		qWarning( "-prefix option requires path argument" );
		exit( 2 );
	    }
	} else if ( arg == "-no-ide" ) {
	    buildIde = FALSE;
	    configs << "noide";
	} else {
	    qWarning( "Unknown option: %s", qapp.argv()[i] );
	    exit( 1 );
	}
    }

    // these are here in case .qmake.cache is missing
#if defined(QT_MODULE_XML)
    configs.append( "xml" );
#endif
#if defined(QT_MODULE_TABLE)
    configs.append( "table" );
#endif
#if defined(QT_MODULE_SQL)
    configs.append( "sql" );
#endif
#if defined(QT_MODULE_NETWORK)
    configs.append( "network" );
#endif

    qtDir = new QString( getenv( "QTDIR" ) );
    if ( qtDir->isEmpty() ) {
	qWarning( "%s: QTDIR not set", app );
	return 1;
    }
    *qtDir += "/";

    if ( !checkLicense() )
	return 1;

    if( !writeQSConfig( buildIde ) )
	return 1;

    if ( !prefix ) {
	// install qmake feature file
	mkDir( *qtDir + "mkspecs/features" );
	copy( "src/qsa/qsa.prf", *qtDir + "mkspecs/features/qsa.prf" );

	// install header files
	QString headers[] = { "qsa/qsaglobal.h",
			      "qsa/qsobjectfactory.h",
			      "qsa/qswrapperfactory.h",
			      "qsa/qseditor.h",
			      "qsa/qsproject.h",
			      "qsa/qsinterpreter.h",
			      "qsa/qsargument.h",
			      "qsa/qsinputdialogfactory.h",
			      "qsa/qsscript.h",
			      "qsa/qsconfig.h",
			      "ide/qsworkbench.h"
	};
	for ( uint j = 0; j < sizeof(headers)/sizeof(QString); j++ ) {
	    int pos = headers[j].find( '/' );
	    QString target = pos >= 0 ? headers[j].mid( pos+1 ) : headers[j];
	    symLink( "src/" + headers[j], *qtDir + "include/" + target );
	}
    }

    installDocs();

    // make the top level Makefile
    QStringList args;
    args.append( "qmake" );
    if ( !configs.isEmpty() )
	args.append( "CONFIG+=" + configs.join( " " ) );
    if ( !prefix.isEmpty() )
	args.append( "QSA_INSTALL_PREFIX=" + prefix );
    if( !execute( args ) )
	warnings++;

    int retVal = processes > 0 ? qapp.exec() : 0;
    if( retVal || errors ) {
	message( "\nThere were errors during configure!" );
    } else if( warnings ) {
	message( "\nThere were errors during configure, but these"
		 "\ndo not appear to be fatal, so you should still be"
		 "\nable to build QSA."
		 "\nRun your make tool to build QSA." );
    } else {
	message( "\n"
		"Configuration completed successfully\n"
		"Run your make tool to build QSA" );
    }

    return retVal || errors;
}

#include "main.moc"
