/****************************************************************************
** Copyright (C) 2001-2003 Trolltech AS.  All rights reserved.
**
** This file is part of the Qt Script for Applications framework (QSA).
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** Licensees holding a valid QSA Beta Evaluation Version license may use
** this file in accordance with the QSA Beta Evaluation Version License
** Agreement provided with the Software.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/pricing.html or email sales@trolltech.com for
**   information about QSA Commercial License Agreements.
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
*****************************************************************************/

#include <stdlib.h>

static QTextEdit *debugoutput = 0;

void debugMessageOutput( QtMsgType type, const char *msg )
{
    if ( type != QtFatalMsg ) {
	if ( debugoutput )
	    debugoutput->append( QString( msg ) + "\n" );
    } else {
	fprintf( stderr, msg );
	abort();
    }
    qApp->flush();
}


static QIconSet createIconSet( const QString &name, bool disabled = TRUE )
{
    QIconSet ic( QPixmap::fromMimeSource( "" + name ) );
    if ( disabled )
	ic.setPixmap( QPixmap::fromMimeSource( "d_" + name ), QIconSet::Small, QIconSet::Disabled );
    return ic;
}

class ScriptListItem : public QListViewItem
{
    friend class IdeWindow;
public:
    ScriptListItem( QListView *parent, QSScript *script )
	: QListViewItem( parent, script->name() ), qss( script )
    {
	QString fname = ( qss && qss->context() ? "scriptobject.png" : "script.png" );
	setPixmap( 0, QPixmap::fromMimeSource( fname ) );
    }

protected:
private:
    QSScript *qss;
};


void IdeWindow::scriptNew()
{
    bool ok;
    QString qsn =
	QInputDialog::getText( "Input Script Name", "Script &Name: ",
			       QLineEdit::Normal,
			       QString( "Script%1.qs" ).arg( project->scripts().count() ),
			       &ok, this );
    if ( ok && !qsn.isEmpty() ) {
	QSScript *script = project->createScript( qsn );
	if ( !script ) return;
	addPage( script );
	enableEditActions( TRUE );
	enableProjectActions( TRUE );
    }
}


void IdeWindow::scriptImport()
{
    QStringList names = QFileDialog::getOpenFileNames( QString::null, QString::null, this );
    for ( QStringList::Iterator it = names.begin(); it != names.end(); ++it ) {
	QFile file( *it );
	if ( file.open( IO_ReadOnly ) ) {
	    QSScript *script = project->createScript( QFileInfo( file ).fileName(), file.readAll() );
	    if ( !script ) return;
	    addPage( script );
	    enableEditActions( TRUE );
	    enableProjectActions( TRUE );
	}
    }
}


void IdeWindow::scriptExport()
{
    ScriptListItem *item = (ScriptListItem *)projectContainer->scriptsListView->currentItem();
    if ( !item ) return;
    QSScript *script = item->qss;
    QSEditor *editor = project->editor( script );
    if ( editor ) editor->commit();
    QFile file( script->name() ); // ### saves in current dir
    if ( file.open( IO_WriteOnly ) ) { // ### check if the file exists, if not call fileSaveAs()
	QTextStream stream( &file );
	stream << script->code();
    }
}


void IdeWindow::scriptExportAs()
{
    ScriptListItem *item = (ScriptListItem *)projectContainer->scriptsListView->currentItem();
    if ( !item ) return;
    QSScript *script = item->qss;
    QString name = QFileDialog::getSaveFileName( script->name(), QString::null, this );
    QFile file( name );
    if ( file.open( IO_WriteOnly ) ) {
	QSEditor *editor = project->editor( script );
	if ( editor ) editor->commit();
	QTextStream stream( &file );
	stream << script->code();
    }
}


void IdeWindow::scriptExit()
{

}


void IdeWindow::editUndo()
{
    QSEditor *editor = (QSEditor *)tabWidget->currentPage();
    if ( !editor ) return;
    editor->undo();
}


void IdeWindow::editRedo()
{
    QSEditor *editor = (QSEditor *)tabWidget->currentPage();
    if ( !editor ) return;
    editor->redo();
}


void IdeWindow::editCut()
{
    QSEditor *editor = (QSEditor *)tabWidget->currentPage();
    if ( !editor ) return;
    editor->cut();
}


void IdeWindow::editCopy()
{
    QSEditor *editor = (QSEditor *)tabWidget->currentPage();
    if ( !editor ) return;
    editor->copy();
}


void IdeWindow::editPaste()
{
    QSEditor *editor = (QSEditor *)tabWidget->currentPage();
    if ( !editor ) return;
    editor->paste();
}


void IdeWindow::editFind()
{
    QSEditor *editor = (QSEditor *)tabWidget->currentPage();
    if ( !editor ) return;
    QString findString = findText->comboFind->currentText();
    findText->comboFind->insertItem( findString );
    bool caseSensitive = findText->checkCase->isChecked();
    bool wholeWordsOnly = findText->checkWhole->isChecked();
    bool startAtCursor = !findText->checkStart->isChecked();
    bool forward = findText->radioForward->isChecked();
    editor->find( findString, caseSensitive, wholeWordsOnly, forward, startAtCursor );
}


void IdeWindow::editReplace()
{
    editReplace( FALSE );
}


void IdeWindow::editReplaceAll()
{
    editReplace( TRUE );
}


void IdeWindow::editReplace( bool all )
{
    QSEditor *editor = (QSEditor *)tabWidget->currentPage();
    if ( !editor ) return;
    QString findString = replaceText->comboFind->currentText();
    replaceText->comboFind->insertItem( findString );
    QString replaceString = replaceText->comboReplace->currentText();
    replaceText->comboFind->insertItem( replaceString );
    bool caseSensitive = replaceText->checkCase->isChecked();
    bool wholeWordsOnly = replaceText->checkWhole->isChecked();
    bool startAtCursor = !replaceText->checkStart->isChecked();
    bool forward = replaceText->radioForward->isChecked();
    editor->replace( findString, replaceString, caseSensitive,
		     wholeWordsOnly, forward, startAtCursor, all );
}


void IdeWindow::editSelectAll()
{
    QSEditor *editor = (QSEditor *)tabWidget->currentPage();
    editor->selectAll();
}


void IdeWindow::editPreferences()
{
    qsaEditorSyntax->reInit();
    preferencesContainer->show();
}


void IdeWindow::editGotoLine()
{
    QSEditor *editor = (QSEditor *)tabWidget->currentPage();
    gotoLine->hide();
    editor->setFocus();
    int val = gotoLine->spinLine->value();
    int max = editor->textEdit()->lines();
    editor->gotoLine( val > max ? max : val );
}


void IdeWindow::helpIndex()
{

}


void IdeWindow::helpContents()
{

}


void IdeWindow::helpAbout()
{
    QMessageBox box( this );
    box.setText( "<center>""<center><img src=\"splash.png\">"
		 "<p>Version " + QString(QSA_VERSION_STRING) + "</p>"
		 "<p>Copyright (C) 2001-2003 Trolltech AS. All rights reserved.</p>"
		 "</center><p></p>"
//		 "<p>This program is licensed to you under the terms of the GNU General "
//		 "Public License Version 2 as published by the Free Software Foundation. This "
//		 "gives you legal permission to copy, distribute and/or modify this software "
//		 "under certain conditions. For details, see the file 'LICENSE.GPL' that came with "
//		 "this software distribution. If you did not get the file, send email to "
//		 "info@trolltech.com.</p>\n\n<p>The program is provided AS IS with NO WARRANTY "
//		 "OF ANY KIND, INCLUDING THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS "
//		 "FOR A PARTICULAR PURPOSE.</p>"
	);
    box.setCaption( tr( "About QSA Workbench" ) );
    box.setIcon( QMessageBox::NoIcon );
    box.exec();
}


void IdeWindow::addPage( QSScript *s )
{
    Q_ASSERT( s );
    QSEditor *editor = project->createEditor( s, tabWidget, s->name() + "editor" );
    tabWidget->addTab( editor, s->name() );
    int idx = tabWidget->indexOf( editor );
    tabWidget->setCurrentPage( idx );
    enableEditActions( TRUE );
}


void IdeWindow::removePage( QSScript *s )
{
    QSEditor *editor = project->editor( s );
    if ( !editor ) return;
    editor->commit();
    tabWidget->removePage( editor );
    delete editor;
    enableEditActions( tabWidget->count() != 0 );
}


void IdeWindow::removePage()
{
    QSEditor *editor = (QSEditor *)tabWidget->currentPage();
    if ( !editor ) return;
    editor->commit();
    tabWidget->removePage( editor );
    delete editor;
    enableEditActions( tabWidget->count() != 0 );
}


void IdeWindow::showPage( QSScript *s )
{
    QSEditor *editor = project->editor( s );
    int idx = -1;
    if ( editor && ( idx = tabWidget->indexOf( editor ) ) != -1 ) {
	tabWidget->setCurrentPage( idx );
	return;
    }
    addPage( s );
}


void IdeWindow::showFunction( QSScript *s, const QString &f )
{
    showPage( s );
    QSEditor *editor = (QSEditor *)tabWidget->currentPage();
    if ( !editor ) return;
    editor->find( "function " + f, FALSE, TRUE, TRUE, FALSE );
}


void IdeWindow::init()
{
    tabWidget->removePage( tabWidget->page( 0 ) );

    QObject::connect( tabWidget, SIGNAL( currentChanged( QWidget * ) ),
		      this, SLOT( currentTabChanged( QWidget * ) ) );

    projectContainerDock = new QDockWindow( QDockWindow::InDock, this );
    projectContainer = new ProjectContainer( projectContainerDock , 0, FALSE );
    projectContainerDock->setResizeEnabled( TRUE );
    projectContainerDock->setCloseMode( QDockWindow::Always );
    addDockWindow( projectContainerDock, DockLeft );
    projectContainerDock->setWidget( projectContainer );
    projectContainerDock->setCaption( "Project Scripts" );
    projectContainerDock->setFixedExtentWidth( 150 );
    projectContainer->scriptsListView->header()->hide();
    projectContainer->show();

    outputContainerDock = new QDockWindow( QDockWindow::InDock, this );
    outputContainer = new OutputContainer( outputContainerDock , 0, FALSE );
    outputContainerDock->setResizeEnabled( TRUE );
    outputContainerDock->setCloseMode( QDockWindow::Always );
    addDockWindow( outputContainerDock, DockBottom );
    outputContainerDock->setWidget( outputContainer );
    outputContainerDock->setCaption( "Output" );
    outputContainerDock->setFixedExtentHeight( 100 );
    outputContainer->show();

    QObject::connect( projectContainer->scriptsListView,
		      SIGNAL( doubleClicked( QListViewItem *) ),
		      this, SLOT( scriptItemDoubleClicked( QListViewItem *) ) );

    findText = new FindText( this, 0, FALSE );
    QObject::connect( editFindAction, SIGNAL( activated() ), findText, SLOT( show() ) );
    QObject::connect( findText->pushFind, SIGNAL( clicked() ), this, SLOT( editFind() ) );

    replaceText = new ReplaceText( this, 0, FALSE );
    QObject::connect( editReplaceAction, SIGNAL( activated() ), replaceText, SLOT( show() ) );
    QObject::connect( replaceText->pushReplace, SIGNAL( clicked() ), this, SLOT( editReplace() ) );
    QObject::connect( replaceText->pushReplaceAll, SIGNAL( clicked() ), this, SLOT( editReplaceAll() ) );
    gotoLine = new GotoLine( this, 0, FALSE );
    QObject::connect( editGotoLineAction, SIGNAL( activated() ), gotoLine, SLOT( show() ) );
    QObject::connect( gotoLine->pushGoto, SIGNAL( clicked() ), this, SLOT( editGotoLine() ) );

    preferencesContainer = new PreferencesContainer( this, 0 );
    QBoxLayout *preferencesLayout = new QBoxLayout( preferencesContainer->frame, QBoxLayout::Down );
    qsaEditorSyntax = new PreferencesBase( preferencesContainer->frame, "qsaeditor_syntax" );
    preferencesLayout->addWidget( qsaEditorSyntax );
    qsaEditorSyntax->setPath( "/Trolltech/QSAScriptEditor/" );
    QObject::connect( preferencesContainer->pushOk, SIGNAL( clicked() ), this, SLOT( savePreferences() ) );

    debugoutput = outputContainer->textEdit;
    qInstallMsgHandler( debugMessageOutput );

    setIcon( QPixmap::fromMimeSource( "qsa.png" ) );
    setupActionIcons();
}


void IdeWindow::removeScript( QSScript *s )
{
    delete s;
    enableProjectActions( projectContainer->scriptsListView->childCount() > 0 );
}


void IdeWindow::removeScript()
{
    ScriptListItem *item = (ScriptListItem *)projectContainer->scriptsListView->currentItem();
    if ( !item )
	return;
    QSScript *script = item->qss;
    delete script;
    enableProjectActions( projectContainer->scriptsListView->childCount() > 0 );
}


void IdeWindow::projectRun()
{
    hadError = FALSE;
    project->commitEditorContents();
    QStringList functions = project->interpreter()->functions();
    if ( hadError )
	return;
    bool ok = TRUE;
    if ( runFunction.isNull()
	 || runFunction == ""
	 || functions.find( runFunction ) == functions.end() ) {
	runFunction = QInputDialog::getItem( "Call function", "&Function: ", functions,
					     functions.findIndex( runFunction ),
					     FALSE, &ok, this );
    }
    if ( ok && !runFunction.isEmpty() )
	project->interpreter()->call( runFunction );
}


void IdeWindow::projectCall()
{
    hadError = FALSE;
    project->commitEditorContents();
    QStringList functions = project->interpreter()->functions();
    if ( hadError )
	return;
    bool ok = FALSE;
    runFunction = QInputDialog::getItem( "Call function", "&Function: ", functions,
					 functions.findIndex( runFunction ),
					 FALSE, &ok, this );
    if ( ok && !runFunction.isEmpty() )
	project->interpreter()->call( runFunction );
}


void IdeWindow::currentTabChanged( QWidget *w )
{
    QSEditor *editor = (QSEditor*)w;
    ScriptListItem *item = (ScriptListItem *)projectContainer->scriptsListView->findItem( editor->script()->name(), 0 );
    projectContainer->scriptsListView->setCurrentItem( item );
    textChanged();
}


void IdeWindow::projectError( const QString &msg, const QString &file, int line )
{
    hadError = TRUE;
    QString err = "<b><font color=red>Error: </font></b> '"
		  + file + "' : <font color=blue>" + QString::number( line ) +
		  "</font><br>" + "<i>" + msg + "</i>";
    debugoutput->append( "<code>" + err + "</code>" );

    QString trace = project->interpreter()->stackTrace().toString();
    debugoutput->append( "<code>Callstack:<br><ul>"
			 + trace.replace( "\n", "<br>" ) + "</ul></code><br>" );

    QSScript *script = project->script( file );
    if ( script ) {
	showPage( script );
	QSEditor *editor = project->editor( script );
	editor->setErrorMark( line );
    }
}


void IdeWindow::scriptItemDoubleClicked( QListViewItem *i )
{
    ScriptListItem *item = (ScriptListItem *)i;
    showPage( item->qss );
}


void IdeWindow::setupActionIcons()
{
    fileNewAction->setIconSet( createIconSet( "scriptnew.png", FALSE ) );
    fileImportAction->setIconSet( createIconSet( "fileopen.png" ) );
    fileExportAction->setIconSet( createIconSet( "filesave.png" ) );
    editUndoAction->setIconSet( createIconSet( "undo.png" ) );
    editRedoAction->setIconSet( createIconSet( "redo.png" ) );
    editCutAction->setIconSet( createIconSet( "editcut.png" ) );
    editCopyAction->setIconSet( createIconSet( "editcopy.png" ) );
    editPasteAction->setIconSet( createIconSet( "editpaste.png" ) );
    editFindAction->setIconSet( createIconSet( "searchfind.png" ) );
    helpContentsAction->setIconSet( createIconSet( "help.png" ) );
    helpAboutAction->setIconSet( createIconSet( "qsa.png", FALSE ) );
    fileRemoveAction->setIconSet( createIconSet( "editdelete.png" ) );
    projectRunAction->setIconSet( createIconSet( "playprev.png" ) );
    projectCallAction->setIconSet( createIconSet( "play.png" ) );
}


void IdeWindow::enableEditActions( bool enable )
{
    fileExportAction->setEnabled( enable );
    fileExportAsAction->setEnabled( enable );
    fileCloseAction->setEnabled( enable );
    editCutAction->setEnabled( enable );
    editCopyAction->setEnabled( enable );
    editPasteAction->setEnabled( enable );
    editFindAction->setEnabled( enable );
    editReplaceAction->setEnabled( enable );
    editGotoLineAction->setEnabled( enable );
    editSelectAllAction->setEnabled( enable );
}


void IdeWindow::enableProjectActions( bool enable )
{
    fileRemoveAction->setEnabled( enable );
    projectRunAction->setEnabled( enable );
    projectCallAction->setEnabled( enable );
}


void IdeWindow::textChanged()
{
    QSEditor *editor = (QSEditor *)tabWidget->currentPage();
    if ( !editor ) return;
    editUndoAction->setEnabled( editor->isUndoAvailable() );
    editRedoAction->setEnabled( editor->isRedoAvailable() );
}


void IdeWindow::setProject( QSProject *p )
{
    Q_ASSERT( p != 0 );
    project = p;
    QObject::connect( project, SIGNAL( editorTextChanged() ),
		      this, SLOT( textChanged() ) );
    projectChanged();
    QPtrList<QSScript> scripts = project->scripts();
    if ( tabWidget->count() == 0 && scripts.count() )
	addPage( scripts.first() );
    connect( project->interpreter(), SIGNAL( error( const QString &, const QString &, int ) ),
	     this, SLOT( projectError( const QString &, const QString &, int ) ) );
    connect( project, SIGNAL( projectChanged() ), this, SLOT( projectChanged() ) );
    errorMode = project->interpreter()->errorMode();
    project->interpreter()->setErrorMode( QSInterpreter::Nothing );
}


void IdeWindow::savePreferences()
{
    qsaEditorSyntax->save();
    QPtrList<QSEditor> editors = project->editors();
    QSEditor *editor = editors.first();
    while ( editor ) {
	editor->readSettings();
	editor = editors.next();
    }
}


void IdeWindow::projectChanged()
{
    QPtrList<QSScript> scripts = project->scripts();
    projectContainer->scriptsListView->clear();
    for ( QSScript *script = scripts.first(); script; script = scripts.next() ) {
	ScriptListItem *item = new ScriptListItem( projectContainer->scriptsListView, script );
	Q_UNUSED( item );
    }
    enableProjectActions( scripts.count() );
}
