/****************************************************************************
** $Id: quickinterpreter.cpp  beta5   edited May 14 17:34 $
**
** Copyright (C) 2001-2002 Trolltech AS.  All rights reserved.
**
** This file is part of the Qt Script for Applications framework (QSA).
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** Licensees holding a valid QSA Beta Evaluation Version license may use
** this file in accordance with the QSA Beta Evaluation Version License
** Agreement provided with the Software.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/pricing.html or email sales@trolltech.com for
**   information about QSA Commercial License Agreements.
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
*****************************************************************************/

#include "quickinterpreter.h"

#include "quickobjects.h"
#include "quickdebugger.h"
#include "quickdispatchobject.h"
#include "quickbytearrayobject.h"
#include "quickpixmapobject.h"
#include "quickcolorobject.h"
#include "quickfontobject.h"
#include "quickcoordobjects.h"
#include "quickobjects.h"
#include "qsenv.h"
#include "qsfunction.h"
#include "qsfuncref.h"
#include <qobjectlist.h>
#include <qmetaobject.h>
#include <qapplication.h>
#include <qdom.h>
#include <private/qucomextra_p.h>
#include <qdir.h>
#include <qsclass.h>
#include <qsnodes.h>
#include <qvaluevector.h>

static bool eTimers = TRUE;
static int static_id = 0;

////////////////////////////////////////////////////////////////////

static QSObject qsConnectCommon( QSEnv *env,
				 QSObject &arg0,
				 QSObject &arg2,
				 QSWrapperShared *&sendObj,
				 QSWrapperShared *&recObj,
				 const QPtrVector<QObject> *&sendIfaces,
				 const QPtrVector<QObject> *&recIfaces,
				 QString &sig,
				 QString &sl,
				 int &signal_index,
				 QObject *&sender,
				 const QString &func )
{
    const QSList &args = *env->arguments();
    QuickInterpreter *ip = QuickInterpreter::fromEnv( env );
    const QString overloads =
	QString( "Following overloads are possible:\n"
		 "%1( sender : QObject, signal : String, receiver : QObject, slot : String )\n"
		 "%2( sender : QObject, signal : String, function : DeclaredFunction )" ).
	arg( func ).arg( func );

    if ( args.size() < 3 || args.size() > 4 ) {
	QString msg = "No matching overload found. " + overloads;
	return env->throwError( SyntaxError, msg );
    }

    QSWrapperClass *wClass = ip->wrapperClass();
    if ( !args[0].isA( wClass ) ) {
	QString msg = "No matching overload found. 'sender' must be of type QObject but is of type " +
		      args[0].typeName() + "\n" + overloads;
	return env->throwError( TypeError, msg );
    }

    if ( args.size() == 3 && !args[2].isFunction() ) {
	QString msg = "No matching overloads found. Third argument in this overload "
		      "must be of type function but is of type " + args[2].typeName() + "\n" + overloads;
	return env->throwError( TypeError, msg );
    }

    arg0 = args[0];
    arg2 = args[2];
    sendObj = wClass->shared( &arg0 );
    recObj =  arg2.isA( wClass ) ? wClass->shared( &arg2 ) : 0;
    sendIfaces = sendObj->interfaceObjects();
    recIfaces = recObj ? recObj->interfaceObjects() : 0;

    // signal and slot in string representation
    sig = args[1].toString();
    sl = ( args.size()>3 ? args[3] : env->createUndefined() ).toString();

    // find sender and signal
    sender = 0;
    signal_index = -1;
    const char *sigName = sig.ascii();
    int i;
    for ( i = (int)sendIfaces->count()-1; i >= 0; --i ) {
	sender = sendIfaces->at( i );
	signal_index = sender->metaObject()->findSignal( sigName, (i == 0) );
	if ( signal_index > 0 )
	    break;
    }

    if ( signal_index == -1 ) {
	// didn't find the signal- this signature might be in Qt
	// Script syntax, so lets's try to find out if we can connect
	QString sn = sig;
	QString arg = sig.mid( sig.find( '(' ) + 1 );
	arg = arg.left( arg.findRev( ')' ) );
	QStringList args = QStringList::split( ',', arg );
	sn = sig.left( sig.find( '(' ) );
	for ( i = (int)sendIfaces->count()-1; i >= 0; --i ) {
	    sender = sendIfaces->at( i );
	    for ( int j = 0; j < (int)sender->metaObject()->numSignals( TRUE ); ++j ) {
		const QMetaData *md = sender->metaObject()->signal( j, TRUE );
		QString mdn = md->name;
		mdn = mdn.left( mdn.find( '(' ) );
		if ( mdn != sn )
		    continue;
		const QUMethod *method = md->method;
		bool ok = method->count == (int)args.count();
		for ( int k = 0; k < method->count; ++k ) {
		    QUParameter p = method->parameters[k];
		    QString s = *args.at( k );
		    int sep;
		    if ( ( sep = s.find( ':' ) ) != -1 )
			s = s.mid( sep + 1 );
		    s = s.simplifyWhiteSpace();
		    if ( s == p.type->desc() || s == QString( (const char*)p.typeExtra ) )
			continue;
		    if ( s == "Number" &&
			 ( qstrcmp( p.type->desc(), "int" ) == 0 ||
			   qstrcmp( p.type->desc(), "long" ) == 0 ||
			   qstrcmp( p.type->desc(), "double" ) == 0 ||
			   qstrcmp( p.type->desc(), "float" ) == 0 ||
			   qstrcmp( p.type->desc(), "short" ) == 0 ||
			   qstrcmp( p.type->desc(), "uint" ) == 0 ||
			   qstrcmp( p.type->desc(), "ushort" ) == 0 ||
			   qstrcmp( p.type->desc(), "ulong" ) == 0 ||
			   qstrcmp( p.type->desc(), "unsigned int" ) == 0 ||
			   qstrcmp( p.type->desc(), "unsigned short" ) == 0 ||
			   qstrcmp( p.type->desc(), "unsigned long" ) == 0 ) )
			continue;
		    s.prepend( "Q" );
		    if ( s == p.type->desc() || s == QString( (const char*)p.typeExtra ) )
			continue;
		    ok = FALSE;
		    break;
		}
		if ( !ok )
		    continue;
		signal_index = j;
		sig = sender->metaObject()->signal( j, TRUE )->name;
		break;
	    }
	    if ( signal_index != -1 )
		break;
	}
    }

    return QSObject();
}

#define QS_CONNECT_INIT( func ) \
    QSObject arg0; \
    QSObject arg2; \
    QSWrapperShared *sendObj; \
    QSWrapperShared *recObj; \
    const QPtrVector<QObject> *sendIfaces = 0; \
    const QPtrVector<QObject> *recIfaces = 0; \
    QString sig; \
    QString sl; \
    int signal_index; \
    QObject *sender = 0; \
    QSObject ret = qsConnectCommon( env, arg0, arg2, sendObj, recObj, \
				    sendIfaces, recIfaces, sig, sl, signal_index, \
				    sender, #func ); \
    if ( ret.isValid() ) { \
	return ret; }


static QSObject qsConnect( QSEnv *env )
{
    QS_CONNECT_INIT( connect );

    // find receiver and slot
    QObject *receiver = 0;
    int member_index = -1;
    const char *slotName = sl.ascii();
    if ( recIfaces ) {
	for ( int i = (int)recIfaces->count()-1; i >= 0; --i ) {
	    receiver = recIfaces->at( i );
	    member_index = receiver->metaObject()->findSlot( slotName, (i == 0) );
	    if ( member_index >= 0 && signal_index >= 0 ) {
		// regular signal/slot connection
		QObject::connectInternal( sender, signal_index,
					  receiver, QSLOT_CODE, member_index );
		return env->createUndefined();
	    }
	}
    }

    if ( signal_index == -1 ) {
	QString msg = "Can't find signal named " + sig;
	return env->throwError( SyntaxError, msg );
    }

    QuickInterpreter *ip = QuickInterpreter::fromEnv( env );
    if ( recIfaces ) {
	sendObj->setEventHandler( ip, sig, recIfaces->at( 0 ), sl.left( sl.find( '(' ) ) );
    } else {
	QSObject base;
	QString name;
	if ( arg2.isFunction() ) {
	    base = QSFuncRefClass::refBase( arg2 );
	    name = QSFuncRefClass::refMember( arg2 ).name();
	} else {
	    base = arg2;
	    if ( base.isPrimitive() )
		return env->throwError( "Invalid receiver object" );
	    name = env->arg( 3 ).toString();
	    if ( name.endsWith( "()" ) )
		name.truncate( name.length() - 2 );
	}
	sendObj->setEventHandler( ip, sig, 0, name, base );
    }

    return env->createUndefined();
}

static QSObject qsDisconnect( QSEnv *env )
{
    QS_CONNECT_INIT( disconnect );

    // find receiver and slot
    QObject *receiver = 0;
    int member_index = -1;
    const char *slotName = sl.ascii();
    if ( recIfaces ) {
	for ( int i = (int)recIfaces->count()-1; i >= 0; --i ) {
	    receiver = recIfaces->at( i );
	    member_index = receiver->metaObject()->findSlot( slotName, (i == 0) );
	    if ( member_index > 0 && signal_index > 0 ) {
		// regular signal/slot connection
		QObject::disconnectInternal( sender, signal_index, receiver,
					     QSLOT_CODE, member_index );
		return env->createUndefined();
	    }
	}
    }

    if ( signal_index == -1 ) {
	QString msg = "Can't find signal named " + sig;
	return env->throwError( SyntaxError, msg );
    }

    if ( recIfaces ) {
	sendObj->removeEventHandler( sig, recIfaces->at( 0 ), sl.left( sl.find( '(' ) ) );
    } else {
	QSObject base = QSFuncRefClass::refBase( arg2 );
	QSMember member = QSFuncRefClass::refMember( arg2 );
	sendObj->removeEventHandler( sig, 0, member.name(), base );
    }

    return env->createUndefined();
}

////////////////////////////////////////////////////////////////////

QuickInterpreter::QuickInterpreter( bool deb )
    : toplevel( 0 ),
      dispatchObjFactories( 0 )
{
    shuttingDown = FALSE;
    id = static_id++;
    wrapperShared = new QSWrapperSharedList;
    //    wrapperShared->setAutoDelete( TRUE );
    usrDataId = QObject::registerUserData();
    globalObjects.setAutoDelete( TRUE );

    debugger = deb ? new QuickDebugger( this ) : 0;
    init();
}

QuickInterpreter::~QuickInterpreter()
{
    shuttingDown = TRUE;
    stopAllTimers();
    invalidateWrappers();
    delete toplevel;
    delete debugger;
    Q_ASSERT( wrapperShared->isEmpty() );
    delete wrapperShared;
    delete dispatchObjFactories;
}

/*!
  Retrieves a pointer to the interpreter class from an environment
  pointer.
 */

QuickInterpreter *QuickInterpreter::fromEnv( QSEnv *e )
{
    return (QuickInterpreter*)e->engine();
}

void QuickInterpreter::reinit()
{
    QSEngine::init();
}

void QuickInterpreter::enableTimers( bool b )
{
    eTimers = b;
}

bool QuickInterpreter::timersEnabled()
{
    return eTimers;
}

QSArgument QuickInterpreter::convertToArgument( const QSObject &o )
{
    if( !o.isValid() )
	return QSArgument();
    const QSClass *cl = o.objectType();
    if( cl->name() == "QObject" ) {
	QSWrapperShared *shared = (QSWrapperShared*) o.shVal();
	Q_ASSERT( shared->objects.count() > 0 );
	return QSArgument( shared->objects[0] );
    } else if( cl == ptrClass ) {
	Q_ASSERT( ptrClass->pointer( &o ) );
	return QSArgument( ptrClass->pointer( &o ) );
    } else {
	return QSArgument( o.toVariant( QVariant::Invalid ) );
    }
}

QSObject QuickInterpreter::wrap( QObject *o )
{
    if( !o )
	return env()->nullClass()->construct( QSList() );
    QSUserData *udata = (QSUserData*) o->userData( userDataId() );
    // set user data in QObject if it's not there, yet
    if( !udata ) {
	udata = new QSUserData( 0 );
	o->setUserData( userDataId(), udata );
    }
    QSWrapperShared *shared = udata->data();
    const QSWrapperClass *cl;
    if ( shared ) {
	// wrapper is already there, reuse it
	cl = shared->wrapperClass();
    } else {
	// create & remember wrapper
	cl = new QSWrapperClass( wrapperClass() );
	shared = cl->createShared( o );
	shared->setObjectType( QSWrapperShared::GlobalObject );
	env()->registerShared( shared );
	wrapperShared->append( shared );
    }
    shared->setUserData( udata );
    udata->setData( shared );
    shared->ref(); // additional ref by either QObject or QSObject
    QSObject obj( cl );
    obj.setVal( shared ); // no ownership needs to be transferred anymore
    return obj;
}

void QuickInterpreter::init()
{
    if ( !dispatchObjFactories ) {
	dispatchObjFactories = new QPtrList<QuickDispatchObjectFactory>;
    }
    debuggerEngine()->clear();

    staticGlobals.clear();

    staticGlobals << "NaN" << "undefined" << "Infinity" << "Application";

    // add some common objects to the Global object
    QSObject global( env()->globalObject() );
    QSClass *objClass = env()->objectClass();

    wrpClass = new QSWrapperClass( objClass );
    ptrClass = new QSPointerClass( objClass );
    varClass = new QSVariantClass( objClass );
    appClass = new QSApplicationClass( objClass );
    global.put( "Application", appClass->createWritable() );

    pntClass = new QSPointClass( objClass, this );
    registerType( pntClass );
    sizClass = new QSSizeClass( objClass, this );
    registerType( sizClass );
    rctClass = new QSRectClass( objClass, this );
    registerType( rctClass );
    colClass = new QSColorClass( objClass );
    registerType( colClass );
    fntClass = new QSFontClass( objClass );
    registerType( fntClass );
    baClass = new QSByteArrayClass( objClass );
    registerType( baClass );
    pixClass = new QSPixmapClass( objClass, this );
    registerType( pixClass );

    enableDebug(); // adds "debug" function which uses qDebug()
    env()->globalClass()->addMember( "connect", QSMember( qsConnect ) );
    env()->globalClass()->addMember( "disconnect", QSMember( qsDisconnect ) );
    env()->globalClass()->addMember( "startTimer", QSMember( qsStartTimer ) );
    env()->globalClass()->addMember( "killTimer", QSMember( qsKillTimer ) );
    env()->globalClass()->addMember( "killTimers", QSMember( qsKillTimers ) );

    for ( QuickDispatchObjectFactory *f = dispatchObjFactories->first();
	  f; f = dispatchObjFactories->next() ) {
	QMap<QString,QObject*> statDescr = f->staticDescriptors();
	QMap<QString,QString> instDescr = f->instanceDescriptors();
	QValueList<QString> features = instDescr.keys();
	for ( QValueList<QString>::ConstIterator it = features.begin();
	      it != features.end(); ++it ) {
	    QSObject staticInst;
	    if( statDescr.contains( *it ) ) { // has static?
		QObject *sinst = statDescr[ *it ];
		Q_ASSERT( sinst );
		staticInst = wrap( sinst );
		statDescr.remove( *it );
	    }
	    QSObjectConstructor *constr =
		new QSObjectConstructor( objClass, *it );
	    QSFactoryObjectProxy *ptype =
		new QSFactoryObjectProxy( objClass, staticInst, constr );
	    QSObject proxy( ptype );
	    proxy.setValue( constr );
	    env()->globalClass()->addStaticVariableMember( constr->identifier(),
							   proxy,
							   AttributeExecutable );
	}

	for( QMap<QString,QObject*>::ConstIterator sit = statDescr.begin();
	     sit != statDescr.end(); ++sit ) {
	    QSObject staticInst;
	    QObject *sinst = statDescr[ sit.key() ];
	    Q_ASSERT( sinst );
	    staticInst = wrap( sinst );
	    env()->globalClass()->addStaticVariableMember( sit.key(), staticInst, AttributeNone );
	}
    }
}

bool QuickInterpreter::hasTopLevelParent( QObject *o )
{
    for ( QObject *p = toplevel->first(); p; p = toplevel->next() ) {
	QObject *c = p->child( o->name(), o->className() );
	if ( c )
	    return TRUE;
    }
    return FALSE;
}

void QuickInterpreter::addTopLevelObject( QObject *o )
{
    QSEngine::init();
    if ( !toplevel )
	toplevel = new QObjectList;
    if ( toplevel->findRef( o ) != -1 )
	return;
    if ( hasTopLevelParent( o ) )
	return;

    for( QObject *cur = toplevel->first(); cur; cur = toplevel->next() ) {
	if( cur == o ) {
	    return;
	} else if ( cur && o && QString( cur->name() ) == QString( o->name() ) ) {
	    return;
	}
    }


    toplevel->append( o );

    kids.clear();
    if ( !toplevel )
	return;
    QObject *obj = toplevel->first();
    while ( obj ) {
	kids.append( obj->name() );
	obj = toplevel->next();
    }
    connect( o, SIGNAL( destroyed( QObject * ) ), this, SLOT( topLevelDestroyed( QObject * ) ) );

    env()->globalObject().put( o->name(), wrap( o ) );
    staticGlobals << o->name();
}

void QuickInterpreter::setTopLevelObjects( QObjectList *l )
{
    QSEngine::init();
    if( toplevel )
	for( QObject *o = toplevel->first(); o; o = toplevel->next() )
	    disconnect( o, SIGNAL( destroyed( QObject* ) ),
			this, SLOT( topLevelDestroyed( QObject* ) ) );
    delete toplevel;
    toplevel = new QObjectList;

    kids.clear();
    if ( !l ) {
	toplevel->clear();
	return;
    }
    QSObject global( env()->globalObject() );
    QObject *o = l->first();
    while ( o ) {
	if ( hasTopLevelParent( o ) ) {
	    o = l->next();
	    continue;
	}
	kids.append( o->name() );
	connect( o, SIGNAL( destroyed( QObject * ) ),
		 this, SLOT( topLevelDestroyed( QObject * ) ) );
	global.put( o->name(), wrap( o ) );
	staticGlobals << o->name();
	toplevel->append( o );
	o = l->next();
    }
    delete l;
}

void QuickInterpreter::topLevelDestroyed( QObject *o )
{
    toplevel->removeRef( o );
}

bool QuickInterpreter::checkSyntax( const QString &c )
{
    QString code = c + "\n";
    return QSEngine::checkSyntax( code );
}

QSArgument QuickInterpreter::execute( QObject *obj, const QString &c,
				      const QString &name )
{
    QString code = c + "\n";

    int sourceId = debugger ? debugger->freeSourceId() : -1;
    if( !name.isNull() && sourceId >= 0 )
	sourceIdNames[sourceId] = name;

    QObject *_this;
    if ( !obj || obj->inherits( "SourceFile" ) ) {	// put into global context
	_this = 0;
    } else {
	_this = obj;
	addTopLevelObject( obj );
    }
    addSourceId( sourceId, obj );

    QSObject t, oldThis;
    if ( _this ) {
	t = wrap( _this );
	oldThis = env()->thisValue();
	env()->setThisValue( t );
    }

    QSEngine::evaluate( t, code );

    // restore this value
    if ( _this )
	env()->setThisValue( oldThis );

    if (hadError())
	emit runtimeError();

    return convertToArgument( returnValue() );
}

QSArgument QuickInterpreter::call( QSObject ctx, const QString &func,
				 const QSList &args )
{
    if ( shuttingDown )
	return QVariant();

    QSEngine::call( &ctx, func, args );

    if (hadError())
	emit runtimeError();

    return convertToArgument( returnValue() );
}

QSArgument QuickInterpreter::call( QObject *ctx, const QString &func,
				 const QSList &args )
{
    if ( shuttingDown )
	return QVariant();

    QSObject t;
    if ( ctx )
	t = wrap( ctx );

    QSEngine::call( &t, func, args );

    if (hadError())
	emit runtimeError();

    return convertToArgument( returnValue() );
}

QSArgument QuickInterpreter::call( QObject *ctx, const QString &func, const QSArgumentList &args )
{
    QSList l;
    for ( QSArgumentList::ConstIterator it = args.begin(); it != args.end(); ++it ) {
	switch ( (*it).type() ) {
	case QSArgument::Variant:
	    l.append( variantClass()->construct( (*it).variant(), 0 ) );
	    break;
	case QSArgument::QObjectPtr:
	    l.append( wrap( (*it).qobject() ) );
	    break;
	case QSArgument::VoidPointer:
	    qWarning( "QuickInterpreter::call: don't know what to do with a QSArgument::VoidPointer here..." );
	    break;
	default:
	    break;
	}
    }
    return call( ctx, func, l );
}

void QuickInterpreter::invalidateWrappers()
{
    wrapperShared->clear();
}

void QuickInterpreter::clear()
{
    sourceIdNames.clear();
    debuggerEngine()->clear();
    stopAllTimers();
    invalidateWrappers();
    if( toplevel )
	toplevel->clear();
    QSEngine::clear();
    QSEngine::init();
    init();
}

void QuickInterpreter::stop()
{
    stopAllTimers();
    debuggerEngine()->clear();
}

bool QuickInterpreter::hadError() const
{
    return !!QSEngine::errorType();
}

/*!
  Emit warning message \a msg for line \a l.
*/

void QuickInterpreter::warn( const QString &msg, int l )
{
    emit warning( msg, l );
}

void QuickInterpreter::clearSourceIdMap()
{
    sourceIdMap.clear();
}

void QuickInterpreter::addSourceId( int id, QObject *o )
{
    sourceIdMap.insert( id, o );
}

QObject *QuickInterpreter::objectOfSourceId( int id ) const
{
    QMap<int, QObject*>::ConstIterator it = sourceIdMap.find( id );
    if ( it == sourceIdMap.end() )
	return 0;
    return *it;
}

int QuickInterpreter::sourceIdOfObject( QObject *o ) const
{
    for ( QMap<int, QObject*>::ConstIterator it = sourceIdMap.begin(); it != sourceIdMap.end(); ++it ) {
	if ( *it == o )
	    return it.key();
	else if( QString( (*it)->name() ) == o->name() )
	    return it.key();
    }
    return -1;
}

QString QuickInterpreter::nameOfSourceId( int id ) const
{
    QMap<int,QString>::ConstIterator it = sourceIdNames.find( id );
    if ( it == sourceIdNames.end() )
	return QString::null;
    return *it;
}

int QuickInterpreter::sourceIdOfName( const QString &name ) const
{
    for ( QMap<int,QString>::ConstIterator it = sourceIdNames.begin();
	  it != sourceIdNames.end(); ++it ) {
	if ( *it == name )
	    return it.key();
    }
    return -1;
}


void QuickInterpreter::cleanType( QString &type )
{
    type = type.simplifyWhiteSpace();
    if ( type.left( 5 ) == "const" )
	type.remove( 0, 5 );
    if ( type[ (int)type.length() - 1 ] == '&' ||
	 type[ (int)type.length() - 1 ] == '*' )
	type.remove( type.length() - 1, 1 );
    type = type.simplifyWhiteSpace();
    if ( type == "QString" )
	type = "String";
    else if ( type == "int" || type == "uint" ||
	      type == "long" || type == "ulong" ||
	      type == "double" || type == "float" )
	type = "Number";
    else if ( type == "bool" )
	type = "Boolean";
}

void QuickInterpreter::cleanTypeRev( QString &type )
{
    if ( type == "String" )
	type = "QString";
    else if ( type == "Number" )
	type = "double";
    else if ( type == "Boolean" )
	type = "bool";
}

bool QuickInterpreter::queryDispatchObjects( QObject *obj,
					     QPtrVector<QObject> &result )
{
    Q_ASSERT( obj );
    QMetaObject *meta = obj->metaObject();
    result.resize( 1 );
    result.insert( 0, obj );
    while ( meta ) {
	QPtrList<QuickDispatchObjectFactory> ifaces =
	    findDispatchObjectFactory( meta->className() );
	if ( ifaces.isEmpty() ) {
	    meta = meta->superClass();
	    continue;
	}
	for ( QuickDispatchObjectFactory *iface = ifaces.first();
	      iface; iface = ifaces.next() )
	    iface->constructInterface( meta->className(), (void*)obj, result );
	meta = meta->superClass();
    }

    return TRUE;
}

bool QuickInterpreter::queryDispatchObjects( const QCString &name, void *ptr,
					     QPtrVector<QObject> &result )
{
    QPtrList<QuickDispatchObjectFactory> ifaces = findDispatchObjectFactory( name );
    if ( ifaces.isEmpty() )
	return FALSE;
    for ( QuickDispatchObjectFactory *iface = ifaces.first(); iface; iface = ifaces.next() )
	iface->constructInterface( name, ptr, result );
    return TRUE;
}

bool QuickInterpreter::construct( const QString &className,
				  const QValueList<QVariant> &vargs,
				  QPtrVector<QObject> &result )
{
    QPtrList<QuickDispatchObjectFactory> ifaces = findDispatchObjectFactory( className );
    if ( ifaces.isEmpty() ) {
	return FALSE;
    }
    for ( QuickDispatchObjectFactory *iface = ifaces.first(); iface; iface = ifaces.next() )
	iface->constructInstance( className, vargs, result );
    return TRUE;
}

QStringList QuickInterpreter::classes() const
{
    QPtrList<QSClass> clsLst = env()->classes();
    QStringList lst;
    for ( QSClass *cls = clsLst.first(); cls; cls = clsLst.next() ) {
	if ( cls->asClass() )
	    lst << cls->asClass()->identifier();
    }
    return lst;
}

void QuickInterpreter::addDispatchObjectFactory( QuickDispatchObjectFactory *f )
{
    dispatchObjFactories->append( f );
    factoryCache.clear();
}

QPtrList<QuickDispatchObjectFactory> QuickInterpreter::findDispatchObjectFactory( const QString &feature )
{
    QMap<QString, QPtrList<QuickDispatchObjectFactory> >::Iterator it =
	factoryCache.find( feature );
    if ( it != factoryCache.end() )
	return *it;
    QPtrList<QuickDispatchObjectFactory> res;
    for ( QuickDispatchObjectFactory *f = dispatchObjFactories->first();
	  f; f = dispatchObjFactories->next() ) {
	QValueList<QString> l = f->classes();
	if ( l.find( feature ) != l.end() )
	    res.append( f );
    }
    factoryCache.insert( feature, res );
    return res;
}

QSObject QuickInterpreter::object( const QString &name ) const
{
    QSObject g = env()->globalObject();
    Global *global = (Global*)&g; // ### ugly
    QSObject obj;
    if ( name.isEmpty() ) {
	obj = g;
    } else {
	int p = name.findRev( '.' );
	if ( p == -1 )
	    obj = global->get( name );
	else
	    obj = global->qualifiedGet( name.left( p ), name.mid( p + 1 ) );
    }
    return obj;
}

const QSClass *QuickInterpreter::classOf( const QSObject &obj ) const
{
    return obj.isA( env()->typeClass() ) ? obj.classValue() : obj.objectType();
}

QStringList QuickInterpreter::functionsOf( QSObject &obj, bool includeSignature, bool includeNative, bool includeMemberFunctions ) const
{
    const QSClass *objType = obj.objectType();
    QSMemberMap mmap = objType == env()->typeClass() && includeMemberFunctions
		       ? ( (QSTypeClass*) objType )->allMembers( &obj )
		       : objType->members( &obj );

    QMap<int, QString> functions;
    for ( QSMemberMap::Iterator it = mmap.begin(); it != mmap.end(); ++it ) {
	if ( ( (*it).type() == QSMember::ScriptFunction ||
	       includeNative && ( (*it).type() == QSMember::NativeFunction ||
				  (*it).type() == QSMember::NativeMemberFunction ) )
	     && !(*it).isPrivate() ) {
	    QString func = (*it).name();
	    // ### will break with mix of script and C++ functions
	    int idx = (*it).type() == QSMember::ScriptFunction ?
		      (*it).scriptFunction->index() : functions.size();
	    if ( includeSignature ) {
		if ( (*it).type() == QSMember::NativeFunction ||
		     (*it).type() == QSMember::NativeMemberFunction ) {
		    func += "()";
		} else {
		    QSFunctionScopeClass *fsc = (*it).scriptFunction->scopeDefinition();
		    int args = fsc->numArguments();
		    func += "(";
		    if ( args > 0 ) {
			func += " ";
			QSMemberMap *members = fsc->definedMembers();
			QValueVector<QString> vec( args );
			for ( QSMemberMap::ConstIterator ait = members->begin();
			      ait != members->end(); ++ait ) {
			    if ( (*ait).index() < args )
				vec[ (*ait).index() ] = (*ait).name();
			}
			for ( int i = 0; i < args; ++i ) {
			    if ( i > 0 )
				func += ", ";
			    func += vec[i];
			}
			func += " ";
		    }
		    func += ")";
		}
	    }
	    functions[idx] = func;
	}
    }

    // assemble to list sorted by original index
    QStringList lst;
    QMap<int, QString>::const_iterator cit = functions.begin();
    for ( ; cit != functions.end(); ++cit )
	    lst << *cit;

    return lst;
}

QStringList QuickInterpreter::classesOf( QSObject &obj ) const
{
    const QSClass *cls = classOf( obj );
    QStringList lst;
    for ( int i = 0; i < cls->numStaticVariables(); ++i ) {
	QSObject o = cls->staticMember( i );
	if ( o.isA( env()->typeClass() ) && o.classValue()->asClass() )
	    lst << o.classValue()->identifier();
    }
    return lst;
}

QStringList QuickInterpreter::variablesOf( QSObject &obj, bool includeStatic, bool includeCustom,
					   bool includeMemberVariables ) const
{
    const QSClass *objType = obj.objectType();
    QSMemberMap mmap = objType == env()->typeClass() && includeMemberVariables
		       ? ( (QSTypeClass*) objType )->allMembers( &obj )
		       : objType->members( &obj );
    QStringList lst;
    for ( QSMemberMap::Iterator it = mmap.begin(); it != mmap.end(); ++it ) {
	if ( ( (*it).type() == QSMember::Variable ||
	       includeCustom && (*it).type() == QSMember::Custom ) &&
	     (!(*it).isStatic() || includeStatic) &&
	     !(*it).isExecutable() )
	    lst << (*it).name();
    }
    return lst;
}
