/****************************************************************************
** $Id: qsinterpreter.cpp  beta5   edited May 16 17:16 $
**
** Implementation of the QSInterpreter class.
**
** Copyright (C) 2001-2002 Trolltech AS.  All rights reserved.
**
** This file is part of the Qt Script for Applications framework (QSA).
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** Licensees holding a valid QSA Beta Evaluation Version license may use
** this file in accordance with the QSA Beta Evaluation Version License
** Agreement provided with the Software.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/pricing.html or email sales@trolltech.com for
**   information about QSA Commercial License Agreements.
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
*****************************************************************************/

#include "qsinterpreter.h"
#include "qsscript.h"
#include "qsobjectfactory.h"
#include "qswrapperfactory.h"
#include "../kernel/quickdebugger.h"
#include "../kernel/quickinterpreter.h"
#include "../kernel/quickobjects.h"
#include "../kernel/quickdispatchobject.h"
#include "../engine/qsenv.h"
#include "../engine/qsfuncref.h"

#include <qapplication.h>
#include <qmessagebox.h>
#include <qptrlist.h>
#include <qcleanuphandler.h>

#ifdef QSA_EXPIRES
static void qsa_eval_check();
#endif

class QSInterpreterPrivate
{
public:
    QSInterpreterPrivate() :
	project( 0 ),
	interpreter( 0 ),
	dispatchFactories( 0 ),
	errorMode( QSInterpreter::Notify )
    {
    }

    ~QSInterpreterPrivate()
    {
	delete dispatchFactories;
	delete interpreter;
    }

    QSProject *project;
    QuickInterpreter *interpreter;
    QuickDispatchObjectFactory *dispatchFactories;
    QSStackTrace stackTrace;
    QSInterpreter::ErrorMode errorMode;
    QPtrList<QSObjectFactory> objectFactories;
    QPtrList<QSWrapperFactory> wrapperFactories;
};

QuickInterpreter *get_quick_interpreter( QSInterpreter *ip )
{
    return ip->d->interpreter;
}

#define startInterpreter() if( d->project ) d->project->evaluate();
#define stopInterpreter()


/*!
  \class QSStackFrame qsinterpreter.h

  \brief The QSStackFrame class provides information about a function call
  in the stack trace.

  A QSStackFrame object only makes sense when grouped together in a
  QSStackTrace to provide a trace of the execution up to where an
  error occurred.

  The QSStackFrame contains four values: file name, line number,
  function name, and context.
*/


/*!
  \fn QSStackFrame::QSStackFrame( const QString &, const QString &,
                                  int line, QObject * )
  \internal
*/


/*!
  \fn QString QSStackFrame::function() const

  Returns the name of the function that this stack frame describes.
*/


/*!
  \fn QString QSStackFrame::scriptName() const

  Returns the name of the script in which the function is located if
  the script name was specified when the function was evaluated;
  otherwise returns QString::null.
*/


/*!
  \fn int QSStackFrame::line() const

  Returns the line number within the script.
*/


/*!
  \fn QObject *QSStackFrame::context() const

  Returns the context of the function if it was evaluated in the
  context of a QObject; otherwise returns 0.
*/


/*!
  \class QSStackTrace qsinterpreter.h

  \brief The QSStackTrace class describes a stack trace after an error
  has occurred in the interpreter.

  The QSStackTrace contains a list of QSStackFrame objects that together
  describe the call stack of the interpreter when an error has occurred.
*/


/*!
  \class QSInterpreter qsinterpreter.h

  \brief The QSInterpreter class provides the public API for the Qt Script
  for Applications script engine.

  This class (implemented in \c libqsa) provides the functionality
  required to make Qt/C++ applications scriptable with Qt Script.

  For convenience reasons, a single instance of the QSInterpreter class
  exists in an application; it is available as
  QSInterpreter::defaultInterpreter().

  The functions evaluate(), call(), addTransientObject(), and clear()
  provide the basic functionality of the interpreter. Any string containing
  valid Qt Script code can be executed using evaluate(), and any state
  built up during evaluation will be kept between calls. The function
  call() can be used to call script functions from C++. The function
  addTransientObject() will add an object to the interpreter for a
  time. Calling clear() will clear the state of the interpreter and
  remove all transient objects.

  The function checkSyntax() provides syntax checking without having
  to execute the context of the code.

  QSInterpreter provides several functions for script introspection.
  These functions are: classes(), functions(), and variables().

  If an error occurs, e.g. during the execution of a script, the
  error() signal is emitted. The error behavior depends on the
  errorMode() which is set with setErrorMode(). When the interpreter
  stops execution because of an error, the hadError(),
  errorMessage(), and stackTrace() functions can be used to provide error information.

  See the \link qsa.book Manual\endlink for more explanations and
  examples.
*/


/*! \enum QSInterpreter::ErrorMode

  The ErrorMode enum describes what happens when an error occurs
  while parsing or executing script code.

  \value Notify Notifies the user that an error occurred via a message
  box. If the application is running in console mode (QApplication::Tty),
  the notification will be output to stderr.

  \value Nothing Nothing happens.
*/


/*!
  \enum QSInterpreter::ClassFlags

  The ClassFlags enum describes which method should be used for class introspection.

  \value AllClasses All available classes in the interpreter, including those
  declared in QObject contexts are matched.

  \value GlobalClasses Only classes declared in the global scope are
  matched.
*/


/*!
  \enum QSInterpreter::FunctionFlags

  The FunctionFlags enum describes matching rules and formatting for
  function introspection.

  \value FunctionNames Returns function names only.

  \value FunctionSignatures Returns the functions with signatures.

  \value IncludeMemberFunctions Matches member functions also, when
  the context to match in is a script class.
*/


/*! \fn bool QSInterpreter::isRunning() const
  Return TRUE if the interpreter is currently evaluating code;
  otherwise returns FALSE.
*/


/*!
  \fn void QSInterpreter::error( const QString &message,
  const QString &scriptName, int lineNumber )

  This signal is emitted if an error occurs when running or parsing a
  script. \a message contains the error message from the interpreter,
  \a scriptName contains the scriptname (if known) of where the error occurred, and \a
  lineNumber contains the line number at which the error occurred.
*/


/*!
  \fn void QSInterpreter::error( const QString &message,
  QObject *context, const QString &scriptName, int lineNumber )

  This signal is emitted if an error occurs when running or parsing a
  script. \a message contains the error message from the interpreter,
  \a context is a pointer to the QObject context in which the error
  occurred or 0, if the context is the global context, \a scriptName
  contains the scriptName (if known) of where the error occurred, and \a lineNumber
  contains the line number at which the error occurred.
*/


/*
  \internal
*/
QSInterpreter::QSInterpreter( QSProject *project, const char *name )
    : QObject( project, name )
{
    d = new QSInterpreterPrivate;
    d->project = project;
    init();
}


/*!
  Constructs a QSInterpreter that runs without a project.

  The \a parent and \a name parameters are passed on to the QObject
  base class.

  There's a default instance accessible with
  QSInterpreter::defaultInterpreter().
*/
QSInterpreter::QSInterpreter( QObject *parent, const char *name )
    : QObject( parent, name )
{
    d = new QSInterpreterPrivate;
    init();
}

/*!
  \internal
*/
void QSInterpreter::init()
{
#ifdef QSA_EXPIRES
    qsa_eval_check();
#endif

    running = FALSE;
    d->interpreter = new QuickInterpreter( TRUE );
    d->dispatchFactories = new QuickDispatchObjectFactory( d->interpreter );
    d->interpreter->addDispatchObjectFactory( d->dispatchFactories );
    connect( d->interpreter, SIGNAL( runtimeError() ),
	     this, SLOT( runtimeError() ) );
}

/*! Destructor
  \internal
*/

QSInterpreter::~QSInterpreter()
{
    for( QSObjectFactory *objs = d->objectFactories.first();
	 objs; objs = d->objectFactories.next() )
	objs->setInterpreter( 0 );

    for( QSWrapperFactory *wraps = d->wrapperFactories.first();
	 wraps; wraps = d->wrapperFactories.next() )
	wraps->setInterpreter( 0 );

    d->objectFactories.setAutoDelete( TRUE );
    d->wrapperFactories.setAutoDelete( TRUE );

    delete d;
}

static QSInterpreter *def_inter = 0;
static QSingleCleanupHandler<QSInterpreter> def_cleanup;

/*!
  Returns the default interpreter.

  The default interpreter runs without a project.

  This function will automatically create the interpreter if it
  doesn't already exist.
 */

QSInterpreter* QSInterpreter::defaultInterpreter()
{
    if ( !def_inter ) {
	def_inter = new QSInterpreter();
	def_cleanup.set( &def_inter );
    }
    return def_inter;
}


/*!
  \property QSInterpreter::errorMode

  \brief What occurs when there is an error:
*/

void QSInterpreter::setErrorMode( ErrorMode m )
{
    d->errorMode = m;
}

QSInterpreter::ErrorMode QSInterpreter::errorMode() const
{
    return d->errorMode;
}


/*!
  Adds the object \a object to the interpreter.

  Transient objects added to the interpreter are not persistent,
  meaning that when the interpreter is cleared, they are also removed.
  Also, when the project is re-evaluated the interpreter is
  cleared and these objects are removed.

  Use QSProject::addObject() to add persistent objects to the
  interpreter.

  \sa QSProject::addObject()
*/
void QSInterpreter::addTransientObject( QObject *object )
{
    d->interpreter->addTopLevelObject( object );
}


/*!
  Executes the string of Qt Script in \a code and returns any value
  produced by that \a code.

  This function executes the code passed in as \a code. The code can
  use and reference code (functions, classes, variables, etc.) which
  have been passed to this function previously or which are defined in
  the current project, if present. Also, application objects which have
  been added via addObject() can be accessed.

  If \a context is 0 (default), the code is executed as global
  code. If a \a context is given, the code is executed in the context
  of that object.

  Interpreters that belong to a project are subject to re-evaluation,
  since the code which has been passed previously into evaluate() gets
  lost when calling one of these functions. This happens when the project
  or the scripts in it are modified.

  \a scriptName can be used for error reporting and debugging.
*/
QSArgument QSInterpreter::evaluate( const QString &code, QObject *context,
				    const QString &scriptName )
{
    running = TRUE;
    startInterpreter();
    QSArgument v = d->interpreter->execute( context, code, scriptName );
    running = FALSE;
    stopInterpreter();
    return v;
}


/*!
  Calls the function \a function with the given \a arguments. The
  arguments are first converted into Qt Script datatypes.

  Functions which were passed to evaluate() in previous calls or which
  are defined in the current project, can be called from this
  function.

  If \a context is 0 (default), the function is called in the global
  scope. If a \a context is given, the function is called in the scope
  of that object.

  Interpreters that belong to a project are subject to re-evaluation,
  since the code which has been passed previously into evaluate() gets
  lost when calling one of these functions. This happens when the project
  or the scripts in it are modified.
*/
QSArgument QSInterpreter::call( const QString &function,
				    const QSArgumentList &arguments,
				    QObject *context )
{
    running = TRUE;
    startInterpreter();
    QSArgument v = d->interpreter->call( context, function, arguments );
    running = FALSE;
    stopInterpreter();
    return v;
}


/*!
  Checks whether the script code \a code is free of syntax errors. Returns TRUE if the code is free of syntax errors; otherwise returns FALSE.
*/
bool QSInterpreter::checkSyntax( const QString &code )
{
    return d->interpreter->checkSyntax( code );
}


/*!
  Returns the current execution context of the interpreter. This is
  either a QObject pointer or 0.
*/
QObject *QSInterpreter::currentContext() const
{
    QSObject o = d->interpreter->env()->currentScope();
    if ( o.isA( d->interpreter->wrapperClass() ) )
	return d->interpreter->wrapperClass()->shared( &o )->objects[0];
    return 0;
}


/*!
  Returns the current project of the interpreter or 0 if there is
  no project.
*/
QSProject *QSInterpreter::project() const
{
    return d->project;
}


/*! \internal */
QuickInterpreter *QSInterpreter::interpreter() const
{
    return d->interpreter;
}


/*!
  Returns the list of objects currently available for scripting in
  this interpreter.
*/
QObjectList QSInterpreter::presentObjects() const
{
    startInterpreter();
    QObjectList *lst = interpreter()->topLevelObjects();
    stopInterpreter();
    return *lst;
}


/*!
  Clears the state of the interpreter.

  When the interpreter is cleared, all declarations parsed using the
  function QSInterpreter::evaluate() are removed. The state of all
  variables will also be cleared.

  Clearing the interpreter will also remove the non-persistant
  objects. Non persistant objects are those added with the function
  QSInterpreter::addObject() or by evaluating code in the context of
  a QObject using QSInterpreter::evaluate();

  This function does not clear persistent application objects added
  by the function QSProject::addObject().
*/
void QSInterpreter::clear()
{
    if( running ) {
#if defined( QT_RANGE_CHECK )
	qWarning( "QSinterpreter::clear(), cannot clear while executing" );
#endif
	return;
    }
    d->interpreter->clear();
    if( d->project ) {
	d->project->invalidateProject();
    }
}


/*!
  Returns all the functions in global context.

  If \a flags does not include FunctionSignatures (default), only the
  function names are returned. If \a flags does include
  FunctionSignatures, then each function name returned will be on the
  following form:

  \code
  functionName( typeOfArg1, typeOfArg2, ... )
  \endcode

  \sa classes(), variables()
*/
QStringList QSInterpreter::functions( FunctionFlags flags ) const
{
    return functions( QString::null, flags );
}


/*!
  Returns all script functions in the context \a context (this can
  be e.g. a class or form). If \a context is empty, the functions of
  the global context (global functions) are returned.

  \a context can be fully qualified.

  If \a flags does not include FunctionSignatures (default), only the
  function names are returned. If \a flags does include
  FunctionSignatures, then each function name returned will be on the
  following form:

  \code
  functionName( typeOfArg1, typeOfArg2, ... )
  \endcode

  If \a flags includes IncludeMemberFunctions and \a context represents
  a class declared in script, this function will return all static and
  non static functions; otherwise only returns static functions.

  \sa classes(), variables()
*/
QStringList QSInterpreter::functions( const QString &context,
				      uint flags ) const
{
    startInterpreter();
    QSObject obj = d->interpreter->object( context );
    QStringList retVal =
	d->interpreter->functionsOf( obj,
				     flags & FunctionSignatures,
				     FALSE,
				     flags & IncludeMemberFunctions );
    stopInterpreter();
    return retVal;
}


/*!
  Returns all script functions which have been defined in the
  context \a context.

  If \a flags does not include FunctionSignatures (default), only the
  function names are returned. If \a flags does include
  FunctionSignatures, then each function name returned will be on the
  following form:

  \code
  functionName( typeOfArg1, typeOfArg2, ... )
  \endcode

  \sa classes(), variables()
*/
QStringList QSInterpreter::functions( QObject *context,
				      FunctionFlags flags ) const
{
    if( !context ) {
#if defined( QT_RANGE_CHECK )
	qWarning( "QSInterpreter::functions: context is null" );
#endif
	return QStringList();
    }
    startInterpreter();
    QStringList lst;
    QSObject obj = d->interpreter->wrap( context );
    if ( !obj.isUndefined() )
	lst = d->interpreter->functionsOf( obj, flags & FunctionSignatures );
    stopInterpreter();
    return lst;
}


/*!
  Returns the classes in the interpreter.

  If \a flags is AllClasses (default), all classes in the
  interpreter are returned, including those declared in object contexts. If
  \a flags is GlobalClasses, only classes declared in the global
  context are returned.

  \sa functions(), variables()
*/
QStringList QSInterpreter::classes( ClassFlags flags ) const
{
    startInterpreter();
    QStringList lst;
    if( flags == AllClasses ) {
	lst = d->interpreter->classes();
    } else {
	QSObject obj = d->interpreter->object( QString::null );
	lst = d->interpreter->classesOf( obj );
    }
    stopInterpreter();
    return lst;
}


/*!
  Returns all classes declared in the fully qualified
  context \a context.

  \sa functions(), variables()
*/
QStringList QSInterpreter::classes( const QString &context ) const
{
    startInterpreter();
    QSObject obj = d->interpreter->object( context );
    QStringList lst = d->interpreter->classesOf( obj );
    stopInterpreter();
    return lst;
}


/*!
  Returns all classes declared in the context \a context.

  \sa functions(), variables()
*/
QStringList QSInterpreter::classes( QObject *context ) const
{
    if( !context ) {
#if defined( QT_RANGE_CHECK )
	qWarning( "QSInterpreter::classesOf: context is null" );
#endif
	return QStringList();
    }
    startInterpreter();
    QSObject obj = d->interpreter->wrap( context );
    QStringList lst;
    if ( !obj.isUndefined() )
	lst = d->interpreter->classesOf( obj );
    stopInterpreter();
    return lst;
}


/*!
  Returns all variables declared in the fully qualified
  context \a context.

  \sa functions(), classes()
*/
QStringList QSInterpreter::variables( const QString &context ) const
{
    startInterpreter();
    QSObject obj = d->interpreter->object( context );
    QStringList lst = d->interpreter->variablesOf( obj );
    stopInterpreter();
    return lst;
}


/*!
  Returns all variables declared in the contenxt \a context.

  \sa functions(), classes()
*/
QStringList QSInterpreter::variables( QObject *context ) const
{
    if( !context ) {
#if defined( QT_RANGE_CHECK )
	qWarning( "QSInterpreter::variablesOf: context is null" );
#endif
	return QStringList();
    }
    startInterpreter();
    QSObject obj = d->interpreter->wrap( context );
    QStringList lst;
    if ( !obj.isUndefined() )
	lst = d->interpreter->variablesOf( obj );
    stopInterpreter();
    return lst;
}


/*!
  Returns all variables declared in the global context.

  \sa functions(), classes()
*/
QStringList QSInterpreter::variables() const
{
    return variables( QString::null );
}


/*!
  \internal
*/
void QSInterpreter::runtimeError()
{
    int l = d->interpreter->debuggerEngine()->lineNumber();
    if ( l < 0 )
	l = d->interpreter->errorLines().first();
    runtimeError( d->interpreter->errorMessages().first(),
		  d->interpreter->nameOfSourceId( d->interpreter->debuggerEngine()->sourceId() ),
		  l );
}


/*!
  \internal
*/
void QSInterpreter::runtimeError( const QString &message,
				  const QString &scriptName,
				  int lineNumber )
{
    emit error( message, scriptName, lineNumber );
    QObject *ctx = d->interpreter->objectOfSourceId( d->interpreter->debuggerEngine()->sourceId() );
    emit error( message, ctx, scriptName, lineNumber );

    if( errorMode() == Notify ) {
	if( qApp->type() == QApplication::Tty ) {
	    qDebug( "Error in script: '%s', line: %d\n  %s\n",
		    scriptName.latin1(), lineNumber, message.latin1() );
	} else {
	    QMessageBox::critical( qApp->mainWidget(), "Error",
				   "The following error occurred in "
				   "line <b>" + QString::number( lineNumber ) +
				   "</b> of  <b>" + scriptName +
				   "</b> while executing "
				   "the script:<pre><font color=red>" +
				   message + "</font></pre>" );
	}
    }
}


/*!
    Informs the interpreter that an error has occurred. The error is
    treated like a normal Qt Script error. The error message is passed
    in \a message.
*/
void QSInterpreter::throwError( const QString &message )
{
    d->interpreter->env()->throwError( message );
}


/*!
  Adds the object factory \a factory to the interpreter.

  Using this function will clear the state of the interpreter.
*/
void QSInterpreter::addObjectFactory( QSObjectFactory *factory )
{
    factory->setInterpreter( this );
    if( factory->interpreter() != this ) {
#if defined( QT_CHECK_STATE )
	qWarning( "QSInterpreter::addObjectFactory(), failed to set interpreter\n" );
#endif
	return;
    }
    d->dispatchFactories->addObjectFactory( factory );
    d->objectFactories.append( factory );
    if( d->project )
	d->project->invalidateProject();
    else
	clear();
}


/*!
  \internal
 */
void QSInterpreter::removeObjectFactory( QSObjectFactory *factory )
{
    d->dispatchFactories->removeObjectFactory( factory );
    d->objectFactories.removeRef( factory );
}


/*!
  Adds the wrapper factory \a factory to the interpreter.

  Using this function will clear the state of the interpreter.
*/
void QSInterpreter::addWrapperFactory( QSWrapperFactory *factory )
{
    factory->setInterpreter( this );
    if( factory->interpreter() != this ) {
#if defined( QT_CHECK_STATE )
	qWarning( "QSInterpreter::addObjectFactory(), failed to set interpreter\n" );
#endif
	return;
    }
    d->dispatchFactories->addWrapperFactory( factory );
    d->wrapperFactories.append( factory );
    if( d->project )
	d->project->invalidateProject();
    else
	clear();
}


/*!
  \internal
*/
void QSInterpreter::removeWrapperFactory( QSWrapperFactory *factory )
{
    d->dispatchFactories->removeWrapperFactory( factory );
    d->wrapperFactories.removeRef( factory );
}


/*!
  Returns the error message that was last reported if there was an
  error; otherwise returns QString::null
*/
QString QSInterpreter::errorMessage() const
{
    return hadError()
	? d->interpreter->errorMessages().first()
	: QString::null ;
}


/*!
  Returns the stack trace describing the call stack of the last
  reported error if there was an error; otherwise returns an empty stack
  trace.
*/
QSStackTrace QSInterpreter::stackTrace() const
{
    QSStackTrace st;
    if( !hadError() )
	return st;

    QValueList<QuickDebuggerStackFrame> qtrace = d->interpreter->debuggerEngine()->backtrace();
    for( QValueList<QuickDebuggerStackFrame>::ConstIterator it = qtrace.begin();
	 it != qtrace.end(); ++it ) {
	QSStackFrame frame( (*it).function,
			    d->interpreter->nameOfSourceId( (*it).sourceId ),
			    (*it).line,
			    d->interpreter->objectOfSourceId( (*it).sourceId )
			    );
	st << frame;
    }
    return st;
}


/*!
  Returns TRUE if the interpreter had an error during the last
  execution; otherwise returns FALSE.
*/
bool QSInterpreter::hadError() const
{
    return d->interpreter->hadError();
}


bool operator==( const QSStackFrame &a, const QSStackFrame &b )
{
    return a.line() == b.line() &&
       a.function() == b.function() &&
       a.scriptName() == b.scriptName() &&
       a.context() == b.context();
}


/*!
  Adds the Qt Script function \a qtscriptFunction (fully qualified)
  as a transient signal handler for the C++ signal \a signal of the
  object \a sender.

  Example:
  \code
  interpreter->addTransientSignalHandler( myButton, SIGNAL( clicked() ), "classA.obj.calculate" );
  \endcode

  \sa removeTransientSignalHandler()
*/
void QSInterpreter::addTransientSignalHandler( QObject *sender,
				  const char *signal,
				  const char *qtscriptFunction )
{
    QuickInterpreter *ip = interpreter();
    QString func( qtscriptFunction );
    func.left( func.find( '(' ) );
    QSObject senderObj = ip->wrap( sender );

    QSObject obj = ip->object( func );
    if( !obj.isFunction() ) {
	qDebug( "QSInterpreter::addTransientSignalHandler(): '%s' not a function",
		qtscriptFunction );
	return;
    }

    QSObject base = QSFuncRefClass::refBase( obj );
    QSMember member = QSFuncRefClass::refMember( obj );
    QSWrapperShared *sh = ip->wrapperClass()->shared( &senderObj );
    sh->setEventHandler( ip, signal + 1, 0, member.name(), base );
}


/*!
  Removes the connection between the signal \a signal of the object
  \a sender and the fully qualified signal handler \a
  qtscriptFunction.

  \sa addTransientSignalHandler()
*/
void QSInterpreter::removeTransientSignalHandler( QObject *sender, const char *signal,
				     const char *qtscriptFunction )
{
    QuickInterpreter *ip = interpreter();
    QString func( qtscriptFunction );
    func.left( func.find( '(' ) );
    QSObject senderObj = ip->wrap( sender );

    QSObject obj = ip->object( func );
    QSObject base = QSFuncRefClass::refBase( obj );
    QSMember member = QSFuncRefClass::refMember( obj );
    QSWrapperShared *sh = ip->wrapperClass()->shared( &senderObj );
    sh->removeEventHandler( signal + 1, 0, member.name(), base );
}



/*!
  Convenience function for describing the stack trace.

  Each stack frame has the following format

  \code
  function(scriptName:lineNumber) context
  \endcode
 */
QString QSStackTrace::toString() const
{
   QStringList lst;
   const char *ctx;
   QString str;
   for( QSStackTrace::ConstIterator it = begin();
	it != end(); ++it ) {
       if( (*it).context() )
	   ctx = str.sprintf( "%s [%s]",
			      (*it).context()->name(),
			      (*it).context()->metaObject()->className() );
       else
	   ctx = "(global context)";
       lst << QString().sprintf( "  %s(%s:%d) %s",
				 (*it).function().latin1(),
				 (*it).scriptName().isNull()
				 ? "<unknown>"
				 : (*it).scriptName().latin1(),
				 (*it).line(),
				 ctx );
   }
   return lst.join( "\n" );
}


#ifdef QSA_EXPIRES
void qsa_eval_check()
{
    QDate date( QSA_EXPIRES>>16, (QSA_EXPIRES&0xff00)>>8, (QSA_EXPIRES&0xff) );
    if ( date > QDate::currentDate() )
	return;
    QMessageBox::warning( 0,
			  QObject::tr( "End of Evaluation Period" ),
			  QObject::tr( "The evaluation period of QSA has expired.\n\n"
				       "Please check http://www.trolltech.com/products/qsa for updates\n"
				       "or contact sales@trolltech.com for further information" ),
			  QMessageBox::Ok,
			  QMessageBox::NoButton );
}
#endif
