/****************************************************************************
** $Id: qsproject.cpp  beta5   edited May 12 16:27 $
**
** Definition of the QSProject class.
**
** Copyright (C) 2001-2003 Trolltech AS.  All rights reserved.
**
** This file is part of the Qt Script for Applications framework (QSA).
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** Licensees holding a valid QSA Beta Evaluation Version license may use
** this file in accordance with the QSA Beta Evaluation Version License
** Agreement provided with the Software.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/pricing.html or email sales@trolltech.com for
**   information about QSA Commercial License Agreements.
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
*****************************************************************************/

#include "qseditor.h"
#include "qsinterpreter.h"
#include "qsproject.h"

#include "../engine/qsfuncref.h"
#include "../kernel/quickdebugger.h"
#include "../kernel/quickinterpreter.h"
#include "../kernel/quickobjects.h"

#include <qapplication.h>
#include <qbuffer.h>
#include <qdatastream.h>
#include <qdir.h>
#include <qfile.h>
#include <qmessagebox.h>
#include <qstringlist.h>

#include <stdlib.h>

#define getQuickInterpreter() d->interpreter->interpreter()

struct QSSignalHandler
{
    QObject *sender, *receiver;
    QString signal, function;
    bool operator==( const QSSignalHandler &sh ) const {
	return ( sh.sender == sender &&
		 sh.receiver == receiver &&
		 sh.signal == signal &&
		 sh.function == function );
    }
};

class QSProjectPrivate {
public:
    QSProjectPrivate() :
	interpreter( 0 ),
	needsRerun( FALSE )
    {
    }

    QSInterpreter *interpreter;
    QPtrList<QSScript> scripts;
    QPtrList<QSEditor> editors;
    QObjectList objects;
    QValueList<QSSignalHandler> signalHandlers;
    QString loadName;
    uint needsRerun : 1;
    uint scriptsModified : 1;
};

/*!
  \class QSProject qsproject.h

  \brief The QSProject class provides project management over scripts
  written in Qt Script for Applications.

  The project's interpreter can be accessed through the function
  interpreter().

  Scripts are encapsulated as instances of the class QSScript. QSScript
  objects are created by using the createScript() functions. It is
  possible to query the project for scripts using the functions scripts(),
  scriptNames(), or script().

  It is possible to associate a QSScript with a context object,
  meaning that the script will be evaluated in the context of that
  object. If a script and an object have the same name, they are
  grouped together.

  Objects are added to the project with addObject() and removed from
  the project with removeObject(). To query the project's objects, use the
  object() and objects() functions.  Note that objects added to the
  project become persistant, meaning that when the interpreter is
  cleared, they will still be scriptable. This is in contrast to
  objects added to the interpreter through the
  QSInterpreter::addTransientObject() function, which will not be available
  after the interpreter is cleared.

  A project can be stored on disk and loaded again using the load()
  and save() functions. It is also possible to write the project to a
  data buffer that can be used in conjunction with other applications
  by using the loadFromData() and saveToData() functions.

  The project also manages editors for the scripts it holds. Editors
  can be created using the functions createEditor() and can be queried
  using the editors(), editor() or activeEditor() functions. When
  text changes in the interpreter, the editorsModified() signal is
  emitted. Before the changes are reflected in the script, the
  commitEditorContents() function must be called. To revert the editors
  to the code in the scripts use revertEditorContents().

  When the editors are committed, the scripts will be out of sync with
  the state of the interpreter. The scriptsModified() signal is
  emitted to indicate this occurrence.

  Several actions in the interpreter will trigger re-evaluation of the
  project; consequently, the interpreter will be cleared and the
  scripts re-evaluated. These actions include modifying scripts and
  removing objects.

  The addSignalHandler() and removeSignalHandler() functions can
  be used to connect and disconnect signals and slots from C++ to
  Qt Script.

  See the \link qsa.book Manual\endlink for more explanations and
  examples.

  \sa QSScript, QSEditor, QSInterpreter
*/


/*!
  \fn QSProject::editorTextChanged()

  This signal is emitted whenever text changes in one of the editors
  that the project currently manages.
*/

/*!
  \fn QSProject::projectChanged()

  \internal

  This signal is emitted whenever the project has changed in such a
  way that it is out of sync with the interpreter and needs to be
  re-evaluated before the interpreter is used.
*/

/*!
  \fn QSProject::projectEvaluated()

  \internal

  This signal is emitted whenever the project has been evaluated and
  the interpreter is no longer out of sync with the scripts in the
  project.
*/


/*!
  Constructs a QSProject with parent \a parent and \a name
*/
QSProject::QSProject( QObject *parent, const char *name )
    : QObject( parent, name )
{
    d = new QSProjectPrivate;
    d->interpreter = new QSInterpreter( this );
    connect( this, SIGNAL( projectChanged() ),
	     this, SLOT( invalidateProject() ) );
}


/*!
  \internal
*/
QSProject::~QSProject()
{
    delete d;
}


/*!
  Reads and parses the Qt Script for Applications project contained in
  the file, \a projectFile.

  If an error occurs while running or parsing the script, actions
  according to the errorMode() are taken and the signal error() is emitted.

  After load() is called, functions and classes defined in this
  project can be called and used through the evaluate() and call()
  functions.

  If the project has signal/slot connections, for example, a form that
  connects widget signals to Qt Script slots, then the connections
  are established as part of the open() process.
*/
bool QSProject::load( const QString &projectFile )
{
    QFile file( projectFile );
    d->loadName = projectFile;
    if( file.open( IO_ReadOnly ) ) {
	QDataStream input( &file );
	return loadInternal( &input );
    }
#if defined(QT_CHECK_RANGE)
    qWarning( "Could not open project file" );
#endif
    return FALSE;
}


/*!
  Reads and parses the Qt Script for Applications project from \a
  projectData.

  If an error occurs while running or parsing the script, actions
  according to the errorMode() are taken and the signal error() is
  emitted.

  After loadFromData() is called, functions and classes defined in this
  project can be called and used through the evaluate() and call()
  functions.

  If the project has signal/slot connections, for example, a form that
  connects widget signals to Qt Script slots, then the connections
  are established as part of the process.
*/
bool QSProject::loadFromData( QByteArray projectData )
{
    QDataStream stream( projectData, IO_ReadOnly );
    return loadInternal( &stream );
}


/*!
  Saves the scripts project to the file \a projectFile

  The default for \a projectFile is QString::null, in which
  case the filename used during load was used.

  \sa load();
*/
bool QSProject::save( const QString &projectFile )
{
    QString fileName = projectFile.isNull() ? d->loadName : projectFile;
    QFile file( fileName );
    if( !file.open( IO_WriteOnly ) ) {
#if defined(QT_CHECK_RANGE)
	qWarning( "QSProject::save(), could not open file for writing" );
#endif
	return FALSE;
    }

    QDataStream output( &file );
    return saveInternal( &output );
}


/*!
  Writes the the whole project to the data block \a data.

  This is useful for keeping the project data with a document. For
  example, you can take the data returned by this function and store
  it in a document's file. Later, when the document is opened, you
  can extract the data from the document and pass it to the
  overload of openFromData() that takes a QByteArray to reconstruct the
  project.

  \sa loadFromData()
*/
bool QSProject::saveToData( QByteArray data )
{
    QDataStream stream( data, IO_WriteOnly );
    return saveInternal( &stream );
}


/*!
  Returns the interpreter which is used to evaluate code
  for this project.
*/
QSInterpreter* QSProject::interpreter() const
{
    return d->interpreter;
}


/*! \fn QSEditor *QSProject::activeEditor() const
  Returns the editor which is currently editing the project if an editor
  is active for this project; otherwise returns 0.
*/
QSEditor* QSProject::activeEditor() const
{
    for ( QSEditor *e = d->editors.first(); e; e = d->editors.next() ) {
	if ( e->isActive() )
	    return e;
    }
    return 0;
}


/*!
  Returns the editor for the script \a script if the editor
  exists; otherwise returns 0.
 */
QSEditor *QSProject::editor( QSScript *script ) const
{
    if( d->scripts.find( script ) < 0 )
	return 0;

    QPtrListIterator<QSEditor> it( d->editors );
    QSEditor *editor = 0;
    QSEditor *current;
    while( ( current = it() ) )
	if( current->script() == script )
	    editor = current;
    return editor;
}


/*!
  Creates a new editor for the script \a script with the parent
  \a parent and name \a name.

  If \a script does not belong to the current project, 0 is returned.

  If an editor already exists for the given script, the existing editor
  is returned.
*/
QSEditor *QSProject::createEditor( QSScript *script, QWidget *parent, const char *name )
{
    if( d->scripts.find( script ) < 0 )
	return 0;

    QSEditor *ed = editor( script );
    if( ed )
	return ed;

    ed = new QSEditor( parent, name );
    ed->setScript( script ); // registerEditor is called from setScript
    return ed;
}


/*!
  Returns a list of all the editors for the scripts in this
  project.
*/
QPtrList<QSEditor> QSProject::editors() const
{
    return d->editors;
}


/*!
  Returns a list of all the scripts in this project.
*/
QPtrList<QSScript> QSProject::scripts() const
{
    return d->scripts;
}


/*!
  Returns the names of all the scripts in this project.
 */
QStringList QSProject::scriptNames() const
{
    QStringList lst;
    QPtrListIterator<QSScript> it( d->scripts );
    QSScript *s;
    while( ( s = it() ) ) {
        lst << s->name();
    }
    return lst;
}


/*!
  Returns the script with the name \a name if it exists in the
  project; otherwise returns 0.
*/
QSScript *QSProject::script( const QString &name ) const
{
    QPtrListIterator<QSScript> it( d->scripts );
    QSScript *s;
    while( ( s = it() ) ) {
	if( s->name() == name )
	    return s;
    }
    return 0;
}


/*!
  Returns the script that has \a context as context if it
  exists in the project; otherwise returns 0.
*/
QSScript *QSProject::script( QObject *context ) const
{
    QPtrListIterator<QSScript> it( d->scripts );
    QSScript *s;
    while( ( s = it() ) ) {
	if( s->context() == context )
	    return s;
    }
    return 0;
}


/*!
  Returns the persistent object in the project with \a name.
*/
QObject *QSProject::object( const QString &name ) const
{
    QObject *object = 0;
    QPtrListIterator<QObject> it( d->objects );
    while( ( object = it() ) ) {
	if( object->name() == name )
	    break;
    }
    return object;
}


/*!
  Returns the list of persistent objects in the project
*/
QObjectList QSProject::objects() const
{
    return d->objects;
}


/*!
  Creates a script with context \a context and source code \a code.

  The context is added to this project's list of objects.

  If a script already exists for this context, 0 is returned.
*/
QSScript *QSProject::createScript( QObject *context, const QString &code )
{
    if( !context ) {
#if defined(QT_CHECK_RANGE)
	qWarning( "QSProject::createScript(), context is null" );
#endif
	return 0;
    }

#if defined(QT_CHECK_RANGE)
    if ( strcmp( context->name(), "unnamed" ) == 0 )
	qWarning( "QSProject::addObject: unnamed object" );
#endif

    return createScriptInternal( context->name(), code, context );
}


/*!
  Creates a script with the name \a name and source code \a code without
  a context.

  If an object exists in the project with the same name as \a name,
  the object is associated with the created script and the script will
  be evaluated in the context of that object.

  If a script already exists with this name, 0 is returned.
*/
QSScript *QSProject::createScript( const QString &name, const QString &code )
{
    if( name.isEmpty() ) {
#if defined(QT_CHECK_RANGE)
	qWarning( "QSProject::createScript(), name cannot be empty!" );
#endif
	return 0;
    }
    return createScriptInternal( name, code, 0 );
}


/*!
  Makes the QObject \a object available to the scripting engine. All child
  objects of \a object are also made available.

  If no object in the parent hierarchy of \a object has been added via
  addObject(), \a object will be made available as a toplevel object to
  the programmer and will be accessible via \c Application.object_name
  (where \c object_name is the value of \e object's QObject::name()
  property).

  If an object in the parent hierarchy of \a object has been made available
  via addObject(), \a object will not be made available as a
  toplevel object. It is then accessible through \c
  parent1.parent2.object_name in the scripting language, given that \c
  parent1 has previously been made available via addObject(). The
  reason to make an object available this way even though it is not
  made available as a toplevel object, is so that code can be added in
  the context of that object.

  Objects added with this function are made persistent. This means that
  when the interpreter is cleared, either by calling QSInterpreter::clear()
  or by modifying the scripts, these objects will still be made available
  to the scripting engine.

  If a script exists in the project that has the same name as \a object,
  the script will be evaluated with \a object as context.

  \warning Every object passed to this function must have a name.

  \sa QSInterpreter::addTransientObject(), removeObject(),
  clearObjects(), object()
*/
void QSProject::addObject( QObject *object )
{
    if ( !object ) {
#if defined(QT_CHECK_RANGE)
	qWarning( "QSProject::addObject(), object is null" );
#endif
	return;
    }
#if defined(QT_CHECK_RANGE)
    if ( strcmp( object->name(), "unnamed" ) == 0 ) {
	qWarning( "QSProject::addObject: unnamed object" );
	return;
    }
#endif
    connect( object, SIGNAL( destroyed() ),
	     this, SLOT( objectDestroyed() ) );
    d->objects.append( object );
    QSScript *s = script( object->name() );
    if( s )
	s->setContext( object );

    getQuickInterpreter()->addTopLevelObject( object );
    emit projectChanged();
}


/*!
  Removes the persistent QObject \a object from the list of
  available objects for scripting.

  Note that only the persistent projects can be removed with
  this function. Objects added to the interpreter using
  QSInterpreter::addTransientObject can not be removed with this
  function.

  If \a object is the context of a script, the script will
  also be removed from the project.

  \sa addObject(), clearObjects(), object()
*/
void QSProject::removeObject( const QObject *object )
{
    if( !object ) {
#if defined(QT_CHECK_RANGE)
	qWarning( "QSProject::removeObject(), object is null" );
#endif
	return;
    }
    QValueList<QSSignalHandler>::Iterator it = d->signalHandlers.begin();
    while ( it != d->signalHandlers.end() ) {
	QValueList<QSSignalHandler>::Iterator it2 = it;
	QSSignalHandler sigh = *it2;
	++it;
	if ( sigh.sender == object || sigh.receiver == object )
	    d->signalHandlers.remove( it2 );
    }
    disconnect( object, SIGNAL( destroyed() ),
		this, SLOT( objectDestroyed() ) );
    d->objects.remove( object );
    QSScript *s = script( object->name() );
    if( s )
	d->scripts.remove( s );

    emit projectChanged();
}


/*!
  Removes all application objects from the list of available
  objects.

  QSProject does not take ownership of the objects, so the objects
  are not cleared.

  \sa addObject(), removeObject(), object(), QSInterpreter::addTransientObject()
*/
void QSProject::clearObjects()
{
    while( d->objects.count() )
	removeObject( d->objects.last() );
    emit projectChanged();
}


/*!
  Adds the Qt Script function \a qtscriptFunction in the context of
  \a receiver as a signal handler for the C++ signal \a signal of the
  object \a sender.

  Example:
  \code
  project->addSignalHandler( myButton, SIGNAL( clicked() ), document, "startCalculation" );
  \endcode

  \sa removeSignalHandler()
*/
void QSProject::addSignalHandler( QObject *sender,
				  const char *signal,
				  QObject *receiver,
				  const char *qtscriptFunction )
{
    evaluate();
    QuickInterpreter *ip = getQuickInterpreter();
    QString func( qtscriptFunction );
    QSObject senderObj = ip->wrap( sender );
    QSWrapperShared *sh = ip->wrapperClass()->shared( &senderObj );
    sh->setEventHandler( ip, signal + 1, receiver, func.left( func.find( '(' ) ) );

    QSSignalHandler sigh;
    sigh.sender = sender;
    sigh.receiver = receiver;
    sigh.signal = signal + 1;
    sigh.function = func.left( func.find( '(' ) );
    d->signalHandlers << sigh;

    connect( sender, SIGNAL( destroyed() ), this, SLOT( objectDestroyed() ) );
    connect( receiver, SIGNAL( destroyed() ), this, SLOT( objectDestroyed() ) );
    emit projectChanged();
}


/*!
  Adds the Qt Script function \a qtscriptFunction (fully qualified)
  as a signal handler for the C++ signal \a signal of the object \a
  sender.

  Example:
  \code
  project->addSignalHandler( myButton, SIGNAL( clicked() ), "a.b.startCalculation" );
  \endcode

  \sa removeSignalHandler()
*/
void QSProject::addSignalHandler( QObject *sender,
				  const char *signal,
				  const char *qtscriptFunction )
{
    evaluate();
    QuickInterpreter *ip = getQuickInterpreter();
    QString func( qtscriptFunction );
    func.left( func.find( '(' ) );
    QSObject senderObj = ip->wrap( sender );

    QSObject obj = ip->object( func );
    if( !obj.isFunction() ) {
	qDebug( "QSProject::addSignalHandler(): '%s' not a function",
		qtscriptFunction );
	return;
    }

    QSObject base = QSFuncRefClass::refBase( obj );
    QSMember member = QSFuncRefClass::refMember( obj );
    QSWrapperShared *sh = ip->wrapperClass()->shared( &senderObj );
    sh->setEventHandler( ip, signal + 1, 0, member.name(), base );

    QSSignalHandler sigh;
    sigh.sender = sender;
    sigh.receiver = 0;
    sigh.signal = signal + 1;
    sigh.function = func;
    d->signalHandlers << sigh;

    connect( sender, SIGNAL( destroyed() ), this, SLOT( objectDestroyed() ) );
    emit projectChanged();
}


/*!
  Removes the connection between the signal \a signal of the object
  \a sender and the signal handler \a qtscriptFunction in the
  context \a receiver.

  \sa addSignalHandler()
*/
void QSProject::removeSignalHandler( QObject *sender, const char *signal,
					 QObject *receiver,
					 const char *qtscriptFunction )
{
    evaluate();
    QString func( qtscriptFunction );
    QuickInterpreter *ip = d->interpreter->interpreter();
    QSObject senderObj = ip->wrap( sender );
    QSWrapperShared *sh = ip->wrapperClass()->shared( &senderObj );
    sh->removeEventHandler( signal + 1, receiver, func.left( func.find( '(' ) ) );
    QSSignalHandler sigh;
    sigh.sender = sender;
    sigh.receiver = receiver;
    sigh.signal = signal + 1;
    sigh.function = func.left( func.find( '(' ) );
    for ( QValueList<QSSignalHandler>::Iterator it = d->signalHandlers.begin();
	  it != d->signalHandlers.end(); ++it ) {
	if ( sigh == *it ) {
	    d->signalHandlers.remove( it );
	    break;
	}
    }
}


/*!
  Removes the connection between the signal \a signal of the object
  \a sender and the fully qualified signal handler \a
  qtscriptFunction.

  \sa addSignalHandler()
*/
void QSProject::removeSignalHandler( QObject *sender, const char *signal,
				     const char *qtscriptFunction )
{
    evaluate();
    QuickInterpreter *ip = getQuickInterpreter();
    QString func( qtscriptFunction );
    func.left( func.find( '(' ) );
    QSObject senderObj = ip->wrap( sender );

    QSObject obj = ip->object( func );
    QSObject base = QSFuncRefClass::refBase( obj );
    QSMember member = QSFuncRefClass::refMember( obj );
    QSWrapperShared *sh = ip->wrapperClass()->shared( &senderObj );
    sh->removeEventHandler( signal + 1, 0, member.name(), base );

    QSSignalHandler sigh;
    sigh.sender = sender;
    sigh.receiver = 0;
    sigh.signal = signal + 1;
    sigh.function = func;
    for ( QValueList<QSSignalHandler>::Iterator it = d->signalHandlers.begin();
	  it != d->signalHandlers.end(); ++it ) {
	if ( sigh == *it ) {
	    d->signalHandlers.remove( it );
	    break;
	}
    }
}


/*!
  Returns TRUE if the project is modified.

  A project enters the modified state when the editors and scripts
  become out of sync. This is the case when a user simply edits a
  script in an editor. Before the interpreter is used, this value
  should be checked and the application should decide to either commit
  or revert the changes in the editor before the project is evaluated.

  \sa commitEditorContents(), revertEditorContents()
*/
bool QSProject::editorsModified() const
{
    QPtrListIterator<QSEditor> it( d->editors );
    QSEditor *current;
    while( ( current = it() ) )
	if( current->isModified() )
	    return TRUE;
    return FALSE;
}


/*!
  Updates the scripts with the content of the editors.

  For each editor that is changed, commit() is called, propagating
  the changes in the editor onto the script it is currently editing.

  After a call to commitEditorContents, the project will no longer be modified

  \sa editorsModified(), revertEditorContents()
 */
void QSProject::commitEditorContents()
{
    QPtrListIterator<QSEditor> it( d->editors );
    QSEditor *current;
    while( ( current = it() ) )
	if( current->isModified() )
	    current->commit();
}


/*!
  Reverts the text in the editors to the state of the scripts.

  After a call to revertEditorContents, the project will no longer
  return TRUE from editorsModified()

  \sa editorsModified(), commitEditorContents()
 */
void QSProject::revertEditorContents()
{
    QPtrListIterator<QSEditor> it( d->editors );
    QSEditor *current;
    while( ( current = it() ) )
	if( current->isModified() )
	    current->revert();
}


/*!
  \internal

  Evaluates all scripts in the project.

  This function clears the state of the interpreter and reevaluates
  all the scripts in the project if needed. The scripts are evaluated
  in their current state, which means that any change made in an
  editor that is not committed will not be taken into account.

  Reevaluation is needed whenever a script has been added, changed or
  removed. Reevaluation is also needed when editors have been committed, persistent objects are added,
  removed or cleared and when signal handlers are modified.
*/
void QSProject::evaluate()
{
    if( !d->needsRerun )
	return;

    QuickInterpreter *ip = getQuickInterpreter();

    ip->clear();
    initObjects();

    QPtrListIterator<QSEditor> eds( d->editors );
    QSEditor *editor;
    while( ( editor = eds() ) ) {
	editor->removeErrorMark();
    }

    QPtrListIterator<QSScript> it( d->scripts );
    QSScript *script;
    while( ( script = it() ) ) {
	ip->execute( script->context(), script->code(), script->name() );
	if( ip->hadError() ) {
	    d->needsRerun = FALSE;
	    return;
	}
    }

    initEventHandlers();

    d->needsRerun = FALSE;
    emit projectEvaluated();
}


/*!
  Returns TRUE if the script has been modified and is out of sync
  with the state in the interpreter; otherwise returns FALSE.
*/
bool QSProject::scriptsModified() const
{
    return d->needsRerun;
}


/*
   \internal
*/
void QSProject::invalidateProject()
{
    d->needsRerun = TRUE;
}


/*! \internal */
void QSProject::initObjects()
{
    QuickInterpreter *ip = getQuickInterpreter();
    for ( QObject *o = d->objects.first(); o;
	  o = d->objects.next() )
	ip->addTopLevelObject( o );
}


/*! \internal */
void QSProject::initEventHandlers()
{
    QuickInterpreter *ip = getQuickInterpreter();
    for ( QValueList<QSSignalHandler>::Iterator it = d->signalHandlers.begin();
	  it != d->signalHandlers.end(); ++it ) {
	QSSignalHandler sigh = *it;
	QSObject senderObj = ip->wrap( sigh.sender );
	QSWrapperShared *sh = ip->wrapperClass()->shared( &senderObj );
	if ( sigh.receiver ) {
	    sh->setEventHandler( ip, sigh.signal, sigh.receiver, sigh.function );
	} else {
	    QSObject obj = ip->object( sigh.function );
	    QSObject base = QSFuncRefClass::refBase( obj );
	    QSMember member = QSFuncRefClass::refMember( obj );
	    sh->setEventHandler( ip, sigh.signal, 0, member.name(), base );
	}
    }
}


/*!
  \internal
*/
bool QSProject::loadInternal( QDataStream *stream )
{
    QString header;
    (*stream) >> header;
    if( header!="QSAv100" ) {
#if defined(QT_CHECK_RANGE)
	qWarning( "QSProject::loadInternal(), datastream is not compatible" );
#endif
	return FALSE;
    }

    d->scripts.clear();
    d->signalHandlers.clear();
    d->editors.setAutoDelete( TRUE );
    d->editors.clear();
    d->editors.setAutoDelete( FALSE );

    int scriptCount;

    (*stream) >> scriptCount;
    QString name, code;
    QObject *context;
    for( int i=0; i<scriptCount; i++ ) {
	(*stream) >> name >> code;
	context = object( name );
	createScriptInternal( name, code, context );
    }

    int shCount;
    (*stream) >> shCount;
    for( int j=0; j<shCount; j++ ) {
	QSSignalHandler handler;
	char *senderName, *receiverName;
	(*stream) >> senderName >> receiverName;
	handler.sender = object( senderName );
	handler.receiver = receiverName ? object( receiverName ) : 0;
	(*stream) >> handler.signal;
	(*stream) >> handler.function;
	if( handler.sender )
	    d->signalHandlers << handler;
	else
	    qWarning( "QSProject::loadInterlan(), unknown sender: " + QString( senderName ) );
    }

    emit projectChanged();
    evaluate();

    return TRUE;
}


/*!
  \internal
*/
bool QSProject::saveInternal( QDataStream *stream )
{
    (*stream) << QString( "QSAv100" );

    (*stream) << d->scripts.count();
    QPtrListIterator<QSScript> it( d->scripts );
    QSScript *script;
    while( ( script = it() ) ) {
	(*stream) << script->name();
	(*stream) << script->code();
    }

    (*stream) << d->signalHandlers.size();
    QValueList<QSSignalHandler>::ConstIterator handlers = d->signalHandlers.begin();
    while( handlers != d->signalHandlers.end() ) {
	(*stream) << (*handlers).sender->name()
		  << ( (*handlers).receiver ? (*handlers).receiver->name() : 0 )
		  << (*handlers).signal
		  << (*handlers).function;
    }

    return TRUE;
}


/*!
  \interal
*/
QSScript *QSProject::createScriptInternal( const QString &name,
					   const QString &code,
					   QObject *context )
{
    Q_ASSERT( !context || context->name() == name );
    if( script( name ) ) {
#if defined(QT_CHECK_RANGE)
	qWarning( "QSProject::createScriptInternal: a script with the name '%s' already exists", name.latin1() );
#endif
	return 0;
    }
    QSScript *s = new QSScript( this, name, code, context );
    d->scripts.append( s );
    if( context && d->objects.find( context ) ) {
	d->objects.append( context );
	connect( context, SIGNAL( destroyed() ),
		 this, SLOT( objectDestroyed() ) );
    }
    connect( s, SIGNAL( codeChanged() ), this, SIGNAL( projectChanged() ) );
    connect( s, SIGNAL( codeChanged() ), this, SLOT( scriptChanged() ) );
    connect( s, SIGNAL( destroyed() ), this, SLOT( objectDestroyed() ) );
    emit projectChanged();
    return s;
}


/*! This function is called when a new QSEditor has been
  created.
  \internal
*/
void QSProject::registerEditor( QSEditor *editor )
{
    if( d->editors.findRef( editor ) >= 0 )
	return;
    d->editors.append( editor );
    connect( editor, SIGNAL( destroyed() ),
	     this, SLOT( objectDestroyed() ) );
    connect( editor, SIGNAL( textChanged() ),
	     this, SIGNAL( editorTextChanged() ) );
}


/*! This function is called when a QSEditor has been
  detroyed.
  \internal
*/
void QSProject::unregisterEditor( QSEditor *editor )
{
    if( d->editors.findRef( editor ) >= 0 )
	return;
    d->editors.removeRef( editor );
    disconnect( editor, SIGNAL( destroyed() ),
		this, SLOT( objectDestroyed() ) );
    disconnect( editor, SIGNAL( textChanged() ),
		this, SIGNAL( editorTextChanged() ) );
}


/*!
  \internal
*/
void QSProject::objectDestroyed()
{
    const QObject *o = QObject::sender();
    if ( d->editors.removeRef( (QSEditor*) o ) )
	return;
    if ( d->scripts.removeRef( (QSScript*) o ) ) {
	emit projectChanged();
	return;
    }
    removeObject( o );
}


/*
  \internal
*/
void QSProject::scriptChanged()
{
    d->scriptsModified = TRUE;
}


#if 0
/*!
  Returns the name of the project.
*/
QString QSProject::projectName() const
{
    return d->projectName;
}
#endif
